# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Panopticon Foundation Model."""

from typing import Any

import timm
import torch
import torch.nn as nn
from torch import Tensor
from torchvision.models._api import Weights, WeightsEnum

from ..samplers.utils import _to_tuple
from .copernicusfm import resize_abs_pos_embed


class PanopticonPE(nn.Module):
    """Defines the Panopticon Patch Embedding."""

    def __init__(
        self,
        attn_dim: int,
        embed_dim: int,
        patch_size: int,
        chnfus_cfg: dict[str, Any] = {},
        img_size: int = 224,
    ) -> None:
        """Initialize a new Panopticon instance.

        Args:
            attn_dim: Embedding dimension on which the channel attention operates.
            embed_dim: Dimension of embeddings that the PanopticonPE outputs.
            patch_size: The patch size.
            chnfus_cfg: Keyword arguments defining the channel attention.
            img_size: Image size.
        """
        super().__init__()

        self.conv3d = Conv3dWrapper(patch_size=patch_size, embed_dim=attn_dim)
        self.chnfus = ChnAttn(**chnfus_cfg, dim=attn_dim)
        self.proj = nn.Linear(attn_dim, embed_dim)

        self.patch_size = _to_tuple(patch_size)
        self.img_size, self.grid_size, self.num_patches = self._init_img_size(img_size)

    def forward(self, x_dict: dict[str, Tensor]) -> Tensor:
        """Forward pass of the model.

        Args:
            x_dict: Dictionary with inputs to the model. Contains keys 'imgs' with
                tensor of shape (B, C, H, W) and 'chn_ids' with tensor of shape (B,C)
                encoding the channel ids.

        Returns:
            Output of shape (B, num_patches, embed_dim).
        """
        x: Tensor = x_dict['imgs']
        chn_ids = x_dict['chn_ids']
        mask = x_dict.get('spec_masks', None)

        x, hp, wp = self.conv3d(x)

        x = self.chnfus(x, chn_ids=chn_ids, mask=mask)  # B,L,D
        x = self.proj(x)

        return x

    def _init_img_size(
        self, img_size: int | tuple[int, int]
    ) -> tuple[tuple[int, int], tuple[int, int], int]:
        """Compute the image size, grid size and number of patches.

        Args:
            img_size: Image size.

        Returns:
            Image size tuple, grid size tuple, and number of patches.
        """
        tuple_img_size = _to_tuple(img_size)
        grid_size = (
            tuple_img_size[0] // self.patch_size[0],
            tuple_img_size[1] // self.patch_size[1],
        )
        num_patches = grid_size[0] * grid_size[1]
        return tuple_img_size, grid_size, num_patches


class Conv3dWrapper(nn.Module):
    """Channel-wise patchification and projection."""

    def __init__(self, patch_size: int, embed_dim: int) -> None:
        """Initialize a conv3d wrapper.

        Args:
            patch_size: Patch size.
            embed_dim: Embedding dimension.
        """
        super().__init__()
        tuple_patch_size = _to_tuple(patch_size)
        patch_CHW = (1, tuple_patch_size[0], tuple_patch_size[1])
        self.conv3d = nn.Conv3d(1, embed_dim, kernel_size=patch_CHW, stride=patch_CHW)

    def forward(self, x: Tensor) -> tuple[Tensor, int, int]:
        """Forward pass.

        Args:
            x: Tensor of shape (B, C, H, W) where B is the batch size.

        Returns:
            The convolved input, number of patches in height, and number of patches in
            width.
        """
        x = self.conv3d(x.unsqueeze(1)).squeeze(1)  # B D C Hp Wp
        hp, wp = x.shape[-2:]
        return x.flatten(-2).permute(0, 2, 3, 1), hp, wp  # B C L D


class ChnAttn(nn.Module):
    """Cross attention over channels with channel embeddings.

    Can reduce any number of channels to a fixed dimension. Inspired by
    https://github.com/microsoft/ClimaX/blob/6d5d354ffb4b91bb684f430b98e8f6f8af7c7f7c/src/climax/arch.py#L185
    """

    def __init__(
        self,
        dim: int,
        chnemb_cfg: dict[str, Any] = {},
        attn_cfg: dict[str, Any] = {},
        layer_norm: bool = False,
    ) -> None:
        """Initialize a channel attention module.

        Args:
            dim: Dimension of the channel attention.
            chnemb_cfg: Key-value pairs for the channel embedding.
            attn_cfg: Key-value pairs for the channel attention.
            layer_norm: Whether to apply layer norm after channel attention.
        """
        super().__init__()

        self.chnemb = ChnEmb(**chnemb_cfg, embed_dim=dim)
        self.query = nn.Parameter(torch.randn(1, 1, dim))
        self.xattn = CrossAttnNoQueryProj(dim=dim, **attn_cfg)

        if layer_norm:
            self.layer_norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor, chn_ids: Tensor, mask: Tensor | None = None) -> Tensor:
        """Forward pass.

        Args:
            x: Image tensor of shape (B, C, L, D).
            chn_ids: Channel IDs tensor of shape (B,C), see ChnEmb.
            mask: Mask tensor of shape (B,C) indicating which channels have been masked
                out.

        Returns:
            Output tensor of shape (B, L, D) independent of the input channel dimension
            C.
        """
        B, C, L, D = x.shape

        # add embeddings
        chn_embs = self.chnemb(chn_ids)  # B,C,D
        x += chn_embs.unsqueeze(2)

        # abstract away L
        x = x.permute(0, 2, 1, 3).flatten(0, 1)  # BL,C,D
        if mask is not None:
            mask = mask.unsqueeze(1).expand(-1, L, -1).flatten(0, 1)  # BL,C

        query = self.query.expand(x.shape[0], -1, -1)  # BL,1,D
        assert query.shape == (x.shape[0], 1, x.shape[-1]), (
            f'Expected query to have shape: {x.shape[0], 1, x.shape[-1]}, but got shape: {query.shape}'
        )

        x = self.xattn(query, x, x, key_padding_mask=mask)
        x = x.reshape(B, L, D)

        if hasattr(self, 'layer_norm'):
            x = self.layer_norm(x)

        return x


class ChnEmb(torch.nn.Module):
    """Computes embeddings from channel IDs."""

    def __init__(
        self, embed_dim: int, use_full_spectra: bool = False, opt_coarsity: int = 1
    ) -> None:
        """Creates embeddings based on the channel IDs.

        Args:
            embed_dim: Embedding dimension.
            use_full_spectra: Whether to additionally to the mean also use the standard
                deviation of optical spectral response (SRF) they are provided.
                This mode only appears in the appendix of the paper.
            opt_coarsity: Define the coarsity of how many nanometers of the mean SRF are
                encoded into the same embedding.
        """
        super().__init__()
        self.embed_dim = embed_dim
        self.use_full_spectra = use_full_spectra
        self.opt_coarsity = opt_coarsity

        dim1 = embed_dim // 3
        dim2 = embed_dim - 2 * dim1
        self.embed_transmit = nn.Parameter(torch.zeros(2, dim1))  # 0:V, 1:H
        self.embed_receive = nn.Parameter(torch.zeros(2, dim1))  # 0:V, 1:H
        self.embed_orbit = nn.Parameter(
            torch.zeros(2, dim2)
        )  # 0:ascending, 1:descending

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Args:
            input: Tensor of shape (B,C) or (B,C,2). If (B,C), we expect the channel
                IDs. If (B,C,2), we expect the channel IDs and the standard deviation
                of the spectral response function (SRF).

        Returns:
            Embeddings of shape (B,C,embed_dim).
        """
        if input.ndim == 2:  # B,C (mus)
            mus = input
        elif input.ndim == 3:  # B,C,2 (mus, sigmas)
            mus = input[:, :, 0]
        sar_indices = mus < 0
        opt_indices = torch.logical_not(sar_indices)
        device = mus.device
        dtype = self.embed_transmit.dtype

        embs = torch.zeros(
            [*list(mus.shape), self.embed_dim], device=device, dtype=dtype
        )

        # Build optical embeddings
        mus[opt_indices] = (mus[opt_indices] // self.opt_coarsity).to(mus.dtype)
        if input.ndim == 2 or not self.use_full_spectra:  # only mus
            embs[opt_indices] = get_1d_sincos_pos_embed_from_grid_torch(
                self.embed_dim, mus[opt_indices].view(-1)
            ).to(dtype)

        elif input.ndim == 3:  # full spectra
            mus_opt = mus[opt_indices]
            sigmas_opt = input[opt_indices][:, 1]
            embs[opt_indices] = get_1d_sincos_ipe_analytical(
                mus_opt, sigmas_opt, self.embed_dim, device
            ).to(dtype)

        # Build SAR embeddings
        transmit = torch.cat(
            [self.embed_transmit[0].repeat(2, 1), self.embed_transmit[1].repeat(2, 1)],
            dim=0,
        ).repeat(3, 1)
        receive = torch.cat(
            [
                self.embed_receive[0].unsqueeze(0),
                self.embed_receive[1].repeat(2, 1),
                self.embed_receive[0].unsqueeze(0),
            ],
            dim=0,
        ).repeat(3, 1)
        orbit = torch.stack(
            [self.embed_orbit.mean(dim=0), self.embed_orbit[0], self.embed_orbit[1]]
        ).repeat_interleave(4, dim=0)
        sar_embs = torch.cat([transmit, receive, orbit], dim=1)

        embs[sar_indices] = sar_embs[(-(mus[sar_indices] + 1)).to(torch.int)]

        return embs


class CrossAttnNoQueryProj(nn.Module):
    """Cross Attention without query projection and final projection."""

    def __init__(self, dim: int, num_heads: int = 8, qkv_bias: bool = False) -> None:
        """Initialize a cross attention module.

        Args:
            dim: Dimension of attention.
            num_heads: Number of heads.
            qkv_bias: Whether to use query, key, and value biases.
        """
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        assert dim % num_heads == 0, 'dim must be divisible by num_heads'
        self.scale = head_dim**-0.5

        self.inproj_q = nn.Identity()  # no projection since query is a parameter itself
        self.inproj_k = nn.Linear(dim, dim, bias=qkv_bias)
        self.inproj_v = nn.Linear(dim, dim, bias=qkv_bias)

    def forward(
        self, q: Tensor, k: Tensor, v: Tensor, key_padding_mask: Tensor | None = None
    ) -> Tensor:
        """Forward pass of the model.

        Args:
            q: Query of shape (B, Nq, D).
            k: Key tensor of shape (B, Nkv, D).
            v: Value tensor of shape (B, Nkv, D).
            key_padding_mask: Key padding mask tensor of shape (B, Nkv).

        Returns:
            Resulting tensor.
        """
        B, Nq, D = q.shape
        q = (
            self.inproj_q(q)
            .reshape(B, Nq, self.num_heads, D // self.num_heads)
            .permute(0, 2, 1, 3)
        )
        q = q * self.scale

        B, Nkv, D = k.shape  # same as v.shape
        k = (
            self.inproj_k(k)
            .reshape(B, Nkv, self.num_heads, D // self.num_heads)
            .permute(0, 2, 1, 3)
        )
        v = (
            self.inproj_v(v)
            .reshape(B, Nkv, self.num_heads, D // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        attn = q @ k.transpose(-2, -1)  # shape: (B, num_heads, Nq, Nkv)
        if key_padding_mask is not None:
            key_padding_mask = key_padding_mask.unsqueeze(1).unsqueeze(
                2
            )  # (B, 1, 1, Nkv)
            attn = attn.masked_fill(key_padding_mask, float('-inf'))

        attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, Nq, D)
        return x


def get_1d_sincos_pos_embed_from_grid_torch(embed_dim: int, pos: Tensor) -> Tensor:
    """Generate standard sin cos positional embeddings.

    Args:
        embed_dim: Output dimension for each position.
        pos: A list of positions to be encoded: size (M,).

    Returns:
        Tensor of embeddings of shape (M,D).
    """
    assert embed_dim % 2 == 0
    omega = torch.arange(embed_dim // 2, dtype=torch.float32, device=pos.device)
    omega /= embed_dim / 2.0
    omega = 1.0 / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = torch.einsum('m,d->md', pos, omega)  # (M, D/2), outer product

    emb_sin = torch.sin(out)  # (M, D/2)
    emb_cos = torch.cos(out)  # (M, D/2)

    emb = torch.cat([emb_sin, emb_cos], dim=1)  # (M, D)
    return emb


def get_1d_sincos_ipe_analytical(
    mu: Tensor, sigma: Tensor, D: int, device: torch.device, temperature: int = 10000
) -> Tensor:
    """Compute the integrated positional embedding (IPE).

    This is only used in the appendix of the paper. You can find further information on
    the motivation and formulas there.

    Args:
        mu: Tensor containing the mus.
        sigma: Tensor containing the sigmas.
        D: Dimension of the embeddings.
        device: Torch device to move the embeddings to.
        temperature: Temperature of embeddings.

    Returns:
        Tensor of embeddings.
    """
    # Create meshgrid for vectorized computation
    d_mesh = torch.arange(D, dtype=torch.float32, device=device)
    mu_mesh = mu.unsqueeze(1).expand(-1, D)
    sigma_mesh = sigma.unsqueeze(1).expand(-1, D)

    # Compute frequencies omega_i
    omega = 1.0 / (temperature ** (2 * d_mesh / D))

    # Compute the Gaussian decay term for each frequency
    # Note: We divide by sigma to normalize similar to how a Gaussian kernel would be normalized
    gaussian_term = torch.exp(-0.5 * (omega.unsqueeze(0) * sigma_mesh) ** 2)

    # Compute sine and cosine terms
    sin_term = torch.sin(omega.unsqueeze(0) * mu_mesh)
    cos_term = torch.cos(omega.unsqueeze(0) * mu_mesh)

    # Combine based on even/odd indices
    IPE = torch.where(
        d_mesh % 2 == 0,
        gaussian_term * sin_term,  # even indices
        gaussian_term * cos_term,
    )  # odd indices

    return IPE


class Panopticon(torch.nn.Module):
    """Panopticon ViT-Base Foundation Model.

    .. versionadded:: 0.7
    """

    def __init__(
        self, attn_dim: int = 2304, embed_dim: int = 768, img_size: int = 224
    ) -> None:
        """Initialize a Panopticon model.

        Args:
            attn_dim: Dimension of channel attention.
            embed_dim: Embedding dimension of backbone.
            img_size: Image size. Panopticon can be initizialized with any image size
                but image size is fixed after initialization. For optimal performance,
                we recommend to use the same image size as used during training.
                For the published weights, this is 224.
        """
        super().__init__()
        dinov2_vit = timm.create_model('vit_base_patch14_dinov2')
        patch_size = 14

        dinov2_vit.patch_embed = PanopticonPE(
            attn_dim=attn_dim,
            embed_dim=embed_dim,
            patch_size=patch_size,
            img_size=img_size,
            chnfus_cfg={'attn_cfg': {'num_heads': 16}},
        )
        dinov2_vit.pos_embed = torch.nn.Parameter(
            torch.randn(1, 1 + (img_size // patch_size) ** 2, embed_dim)
        )

        self.model: nn.Module = dinov2_vit

    def forward(self, x_dict: dict[str, Tensor]) -> Tensor:
        """Forward pass of the model including forward pass through the head.

        Args:
            x_dict: Dictionary with keys:

                * imgs: Input tensor of shape (B, C, H, W).
                * chn_ids: Tensor of shape (B,C) encoding the spectral information
                  of each channel. For optical channels, this is the wavelength in
                  nanometers. For SAR channels, this is a negative integer as outlined
                  in https://github.com/Panopticon-FM/panopticon/blob/main/dinov2/configs/data/satellites/sentinel1.yaml

        Returns:
            Embeddings.
        """
        out: Tensor = self.model.forward(x_dict)
        return out


class Panopticon_Weights(WeightsEnum):  # type: ignore[misc]
    """Panopticon weights.

    .. versionadded:: 0.7
    """

    VIT_BASE14 = Weights(
        url='https://hf.co/lewaldm/panopticon/resolve/c8c2bb9555819e8b2bcedf5b3b00e3bf531554e7/panopticon_vitb14_teacher.pth',
        transforms=None,
        meta={
            'model': 'panopticon_vitb14',
            'publication': 'https://arxiv.org/abs/2503.10845',
            'repo': 'https://github.com/Panopticon-FM/panopticon',
            'ssl_method': 'dinov2+spectral_progressive_pretraining',
        },
    )


def panopticon_vitb14(
    weights: Panopticon_Weights | None = None, img_size: int = 224, **kwargs: int
) -> Panopticon:
    """Panopticon ViT-Base model.

    Panopticon can handle arbitrary optical channel and SAR combinations.
    It can also be initialized with any image size where the image size is
    fixed after initialization. However, we recommend to set 224 in alignment
    with the pretraining. For more information on how to use the model,
    see https://github.com/Panopticon-FM/panopticon?tab=readme-ov-file#using-panopticon.

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/abs/2503.10845

    .. versionadded:: 0.7

    Returns:
        The Panopticon ViT-Base model with the published weights loaded.
    """
    model = Panopticon(img_size=img_size, **kwargs)
    patch_size = 14  # fixed

    if weights:
        state_dict = weights.get_state_dict(progress=True)
        state_dict.pop('mask_token')

        # interpolate positional embeddings (timm==0.9.2) does not support this yet
        state_dict['pos_embed'] = resize_abs_pos_embed(
            state_dict['pos_embed'], img_size // patch_size, 518 // patch_size
        )

        missing_keys, unexpected_keys = model.model.load_state_dict(
            state_dict, strict=True
        )
        assert not missing_keys
        assert not unexpected_keys

    return model



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""DOTA dataset."""

import os
from collections.abc import Callable
from typing import Any, ClassVar, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from matplotlib import patches
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import (
    Path,
    check_integrity,
    download_url,
    extract_archive,
    percentile_normalization,
)


class DOTA(NonGeoDataset):
    """DOTA dataset.

    `DOTA <https://captain-whu.github.io/DOTA/index.html>`__ is a large-scale object
    detection dataset for aerial imagery containing RGB and gray-scale imagery
    from Google Earth, GF-2 and JL-1 satellites as well as additional aerial imagery
    from CycloMedia. There are three versions of the dataset: v1.0, v1.5, and v2.0, where,
    v1.0 and v1.5 have the same images but different annotations,
    and v2.0 extends both the images and annotations with more samples

    Dataset features:

    * 1869 samples in v1.0 and v1.5 and 2423 samples in v2.0
    * multi-class object detection (15 classes in v1.0 and v1.5 and 18 classes in v2.0)
    * horizontal and oriented bounding boxes

    Dataset format:

    * images are three channel PNGs with various pixel sizes
    * annotations are text files with one line per bounding box

    Classes:

    0. plane
    1. ship
    2. storage-tank
    3. baseball-diamond
    4. tennis-court
    5. basketball-court
    6. ground-track-field
    7. harbor
    8. bridge
    9. large-vehicle
    10. small-vehicle
    11. helicopter
    12. roundabout
    13. soccer-ball-field
    14. swimming-pool
    15. container-crane (v1.5+)
    16. airport (v2.0+)
    17. helipad (v2.0+)

    If you use this work in your research, please cite the following papers:

    * https://arxiv.org/abs/2102.12219
    * https://arxiv.org/abs/1711.10398

    .. versionadded:: 0.7
    """

    url = 'https://huggingface.co/datasets/torchgeo/dota/resolve/672e63236622f7da6ee37fca44c50ac368b77cab/{}'

    file_info: ClassVar[dict[str, dict[str, dict[str, dict[str, str]]]]] = {
        'train': {
            'images': {
                '1.0': {
                    'filename': 'dotav1.0_images_train.tar.gz',
                    'md5': '363b472dc3c71e7fa2f4a60223b437ea',
                },
                '1.5': {
                    'filename': 'dotav1.0_images_train.tar.gz',
                    'md5': '363b472dc3c71e7fa2f4a60223b437ea',
                },
                '2.0': {
                    'filename': 'dotav2.0_images_train.tar.gz',
                    'md5': '91ae5212d170330ab9f65ccb6c675763',
                },
            },
            'annotations': {
                '1.0': {
                    'filename': 'dotav1.0_annotations_train.tar.gz',
                    'md5': 'f6788257bcc4d29018344a4128e3734a',
                },
                '1.5': {
                    'filename': 'dotav1.5_annotations_train.tar.gz',
                    'md5': '0da97e5623a87d7bec22e75f6978dbce',
                },
                '2.0': {
                    'filename': 'dotav2.0_annotations_train.tar.gz',
                    'md5': '04d3d626df2203053b7f06581b3b0667',
                },
            },
        },
        'val': {
            'images': {
                '1.0': {
                    'filename': 'dotav1.0_images_val.tar.gz',
                    'md5': '42293219ba61d61c417ae558bbe1f2ba',
                },
                '1.5': {
                    'filename': 'dotav1.0_images_val.tar.gz',
                    'md5': '42293219ba61d61c417ae558bbe1f2ba',
                },
                '2.0': {
                    'filename': 'dotav2.0_images_val.tar.gz',
                    'md5': '737f65edf54b5aa627b3d48b0e253095',
                },
            },
            'annotations': {
                '1.0': {
                    'filename': 'dotav1.0_annotations_val.tar.gz',
                    'md5': '28155c05b1dc3a0f5cb6b9bdfef85a13',
                },
                '1.5': {
                    'filename': 'dotav1.5_annotations_val.tar.gz',
                    'md5': '85bf945788784cf9b4f1c714453178fc',
                },
                '2.0': {
                    'filename': 'dotav2.0_annotations_val.tar.gz',
                    'md5': 'ec53c1dbcfc125d7532bd6a065c647ac',
                },
            },
        },
    }

    sample_df_path = 'samples.csv'

    classes = (
        'plane',
        'ship',
        'storage-tank',
        'baseball-diamond',
        'tennis-court',
        'basketball-court',
        'ground-track-field',
        'harbor',
        'bridge',
        'large-vehicle',
        'small-vehicle',
        'helicopter',
        'roundabout',
        'soccer-ball-field',
        'swimming-pool',
        'container-crane',
        'airport',
        'helipad',
    )

    valid_splits = ('train', 'val')
    valid_versions = ('1.0', '1.5', '2.0')

    valid_orientations = ('horizontal', 'oriented')

    def __init__(
        self,
        root: Path = 'data',
        split: Literal['train', 'val'] = 'train',
        version: Literal['1.0', '1.5', '2.0'] = '2.0',
        bbox_orientation: Literal['horizontal', 'oriented'] = 'oriented',
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new DOTA dataset instance.

        Args:
            root: root directory where dataset can be found
            split: split of the dataset to use, one of ['train', 'val']
            version: version of the dataset to use, one of ['1.0', '1.5', '2.0']
            bbox_orientation: bounding box orientation, one of ['horizontal', 'oriented'], where horizontal
                returnx xyxy format and oriented returns x1y1x2y2x3y3x4y4 format
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if *split*, *version*, or *bbox_orientation* argument are not valid
            DatasetNotFoundError: if dataset is not found or corrupted, and *download* is False
        """
        assert split in self.valid_splits, (
            f"Split '{split}' not supported, use one of {self.valid_splits}"
        )
        assert version in self.valid_versions, (
            f"Version '{version}' not supported, use one of {self.valid_versions}"
        )

        assert bbox_orientation in self.valid_orientations, (
            f'Bounding box orientation must be one of {self.valid_orientations}'
        )

        self.root = root
        self.split = split
        self.version = version
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.bbox_orientation = bbox_orientation

        self._verify()

        self.sample_df = pd.read_csv(os.path.join(self.root, 'samples.csv'))
        self.sample_df['version'] = self.sample_df['version'].astype(str)
        self.sample_df = self.sample_df[self.sample_df['split'] == self.split]
        self.sample_df = self.sample_df[
            self.sample_df['version'] == self.version
        ].reset_index(drop=True)

    def __len__(self) -> int:
        """Return the number of samples in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.sample_df)

    def __getitem__(self, index: int) -> dict[str, Any]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        sample_row = self.sample_df.iloc[index]

        sample = {'image': self._load_image(sample_row['image_path'])}

        boxes, labels = self._load_annotations(sample_row['annotation_path'])

        if self.bbox_orientation == 'horizontal':
            sample['bbox_xyxy'] = boxes
        else:
            sample['bbox'] = boxes
        sample['labels'] = labels

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_image(self, path: str) -> Tensor:
        """Load image.

        Args:
            path: path to image file

        Returns:
            image: image tensor
        """
        image = Image.open(os.path.join(self.root, path)).convert('RGB')
        return torch.from_numpy(np.array(image).transpose(2, 0, 1)).float()

    def _load_annotations(self, path: str) -> tuple[Tensor, Tensor]:
        """Load DOTA annotations from text file.

        Format:
            x1 y1 x2 y2 x3 y3 x4 y4 class difficult

        Some files have 2 header lines that need to be skipped:
            imagesource:GoogleEarth
            gsd:0.146343590398

        Args:
            path: path to annotation file

        Returns:
            tuple of:
                boxes: tensor of shape (N, 8) with coordinates for oriented
                    and (N, 4) for horizontal
                labels: tensor of shape (N,) with class indices
        """
        with open(os.path.join(self.root, path)) as f:
            lines = f.readlines()

        # Skip header if present
        start_idx = 0
        if lines and lines[0].startswith('imagesource'):
            start_idx = 2
        boxes = []
        labels = []

        for line in lines[start_idx:]:
            parts = line.strip().split(' ')

            # Always read 8 coordinates
            coords = [float(p) for p in parts[:8]]
            label = parts[8]

            labels.append(self.classes.index(label))

            if self.bbox_orientation == 'horizontal':
                # Convert to [xmin, ymin, xmax, ymax] format
                x_coords = coords[::2]  # even indices (0,2,4,6)
                y_coords = coords[1::2]  # odd indices (1,3,5,7)
                xmin, xmax = min(x_coords), max(x_coords)
                ymin, ymax = min(y_coords), max(y_coords)
                boxes.append([xmin, ymin, xmax, ymax])
            else:
                boxes.append(coords)

        if not boxes:
            return (
                torch.zeros((0, 4 if self.bbox_orientation == 'horizontal' else 8)),
                torch.zeros(0, dtype=torch.long),
            )
        else:
            return torch.tensor(boxes), torch.tensor(labels)

    def _verify(self) -> None:
        """Verify dataset integrity and download/extract if needed."""
        # check if directories and sample file are present
        required_dirs = [
            os.path.join(self.root, self.split, 'images'),
            os.path.join(
                self.root, self.split, 'annotations', f'version{self.version}'
            ),
            os.path.join(self.root, self.sample_df_path),
        ]
        if all(os.path.exists(d) for d in required_dirs):
            return

        # Check for compressed files, v1.0 and v1.5 have the same images but different annotations
        files_needed = [
            (
                self.file_info[self.split]['images'][self.version]['filename'],
                self.file_info[self.split]['images'][self.version]['md5'],
            ),
            (
                self.file_info[self.split]['annotations'][self.version]['filename'],
                self.file_info[self.split]['annotations'][self.version]['md5'],
            ),
        ]
        # For v2.0, also need v1.0 image files, but only v2 annotations
        if self.version == '2.0':
            files_needed.append(
                (
                    self.file_info[self.split]['images']['1.0']['filename'],
                    self.file_info[self.split]['images']['1.0']['md5'],
                )
            )

        # Check if archives exist and verify checksums if requested
        exists = []
        for filename, md5 in files_needed:
            filepath = os.path.join(self.root, filename)
            if os.path.exists(filepath):
                if self.checksum:
                    if not check_integrity(filepath, md5):
                        raise RuntimeError(f'Archive {filename} corrupted')
                exists.append(True)
                self._extract([(filename, md5)])
            else:
                exists.append(False)

        if all(exists):
            return

        if not self.download:
            raise DatasetNotFoundError(self)

        # also download the metadata file
        self._download(files_needed)
        self._extract(files_needed)

    def _download(self, files_needed: list[tuple[str, str]]) -> None:
        """Download the dataset.

        Args:
            files_needed: list of files to download for the particular version
        """
        for filename, md5 in files_needed:
            if not os.path.exists(os.path.join(self.root, filename)):
                download_url(
                    url=self.url.format(filename),
                    root=self.root,
                    filename=filename,
                    md5=None if not self.checksum else md5,
                )

        if not os.path.exists(os.path.join(self.root, self.sample_df_path)):
            download_url(
                url=self.url.format(self.sample_df_path),
                root=self.root,
                filename=self.sample_df_path,
            )

    def _extract(self, files_needed: list[tuple[str, str]]) -> None:
        """Extract the dataset.

        Args:
            files_needed: list of files to extract for the particular version
        """
        for filename, _ in files_needed:
            filepath = os.path.join(self.root, filename)
            extract_archive(filepath, self.root)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
        box_alpha: float = 0.7,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by __getitem__
            show_titles: flag indicating whether to show titles
            suptitle: optional string to use as a suptitle
            box_alpha: alpha value for boxes

        Returns:
            a matplotlib Figure with the rendered sample
        """
        image = percentile_normalization(sample['image'].permute(1, 2, 0).numpy())
        if self.bbox_orientation == 'horizontal':
            boxes = sample['bbox_xyxy'].cpu().numpy()
        else:
            boxes = sample['bbox'].cpu().numpy()
        labels = sample['labels'].cpu().numpy()

        fig, ax = plt.subplots(figsize=(10, 10))
        ax.imshow(image)
        ax.axis('off')

        # Create color map for classes
        cm = plt.get_cmap('gist_rainbow')

        for box, label_idx in zip(boxes, labels):
            color = cm(label_idx / len(self.classes))
            label = self.classes[label_idx]

            if self.bbox_orientation == 'horizontal':
                # Horizontal box: [xmin, ymin, xmax, ymax]
                x1, y1, x2, y2 = box
                rect = patches.Rectangle(
                    (x1, y1),
                    x2 - x1,
                    y2 - y1,
                    linewidth=2,
                    alpha=box_alpha,
                    linestyle='solid',
                    edgecolor=color,
                    facecolor='none',
                )
                ax.add_patch(rect)
                # Add label above box
                ax.text(
                    x1,
                    y1 - 5,
                    label,
                    color='white',
                    fontsize=8,
                    bbox=dict(facecolor=color, alpha=box_alpha),
                )
            else:
                # Oriented box: [x1,y1,x2,y2,x3,y3,x4,y4]
                vertices = box.reshape(4, 2)
                polygon = patches.Polygon(
                    vertices,
                    linewidth=2,
                    alpha=box_alpha,
                    linestyle='solid',
                    edgecolor=color,
                    facecolor='none',
                )
                ax.add_patch(polygon)
                # Add label at centroid
                centroid_x = vertices[:, 0].mean()
                centroid_y = vertices[:, 1].mean()
                ax.text(
                    centroid_x,
                    centroid_y,
                    label,
                    color='white',
                    fontsize=8,
                    bbox=dict(facecolor=color, alpha=box_alpha),
                    ha='center',
                    va='center',
                )

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Fields Of The World dataset."""

import os
from collections.abc import Callable, Sequence
from typing import ClassVar

import einops
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, array_to_tensor, download_and_extract_archive, extract_archive


class FieldsOfTheWorld(NonGeoDataset):
    """Fields Of The World dataset.

    The `Fields Of The World <https://beta.source.coop/repositories/kerner-lab/fields-of-the-world/>`__
    datataset is a semantic and instance segmentation dataset for delineating field
    boundaries.

    Dataset features:

    * 70462 patches across 24 countries
    * Each country has a train, val, and test split
    * Semantic segmentations masks with and without the field boundary class
    * Instance segmentation masks

    Dataset format:

    * images are four-channel GeoTIFFs with dimension 256x256
    * segmentation masks (both two and three class) are single-channel GeoTIFFs
    * instance masks are single-channel GeoTIFFs

    Dataset classes:

    0. background
    1. field
    2. field-boundary (three-class only)
    3. unlabeled (kenya, rwanda, brazil and india have presence only labels)

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.48550/arXiv.2409.16252

    .. versionadded:: 0.7
    """

    splits = ('train', 'val', 'test')
    targets = ('2-class', '3-class', 'instance')

    valid_countries = (
        'austria',
        'belgium',
        'brazil',
        'cambodia',
        'corsica',
        'croatia',
        'denmark',
        'estonia',
        'finland',
        'france',
        'germany',
        'india',
        'kenya',
        'latvia',
        'lithuania',
        'luxembourg',
        'netherlands',
        'portugal',
        'rwanda',
        'slovakia',
        'slovenia',
        'south_africa',
        'spain',
        'sweden',
        'vietnam',
    )

    base_url = 'https://data.source.coop/kerner-lab/fields-of-the-world-archive/'

    country_to_md5: ClassVar[dict[str, str]] = {
        'austria': '35604e3e3e78b4469e443bc756e19d26',
        'belgium': '111a9048e15391c947bc778e576e99b4',
        'brazil': '2ba96f9f01f37ead1435406c3f2b7c63',
        'cambodia': '581e9b8dae9713e4d03459bcec3c0bd0',
        'corsica': '0b38846063a98a31747fdeaf1ba03980',
        'croatia': 'dc5d33e19ae9e587c97f8f4c9852c87e',
        'denmark': 'ec817210b06351668cacdbd1a8fb9471',
        'estonia': 'b9c89e559e3c7d53a724e7f32ccf88ea',
        'finland': '23f853d6cbaea5a3596d1d38cc27fd65',
        'france': 'f05314f148642ff72d8bea903c01802d',
        'germany': 'd57a7ed203b9cf89c709aab29d687cee',
        'india': '361a688507e2e5cc7ca7138be01a5b80',
        'kenya': '80ca0335b25440379f99b7011dfbdfa2',
        'latvia': '6eeaaa57cdf18f25497f84e854a86d42',
        'lithuania': '0a2f4ab3309633e2de121d936e0763ba',
        'luxembourg': '5a8357eae364cca836b87827b3c6a3d3',
        'netherlands': '3afc61d184aab5c4fd6beaecf2b6c0a9',
        'portugal': '10485b747e1d8c082d33c73d032a7e05',
        'rwanda': '087ce56bbf06b32571ef27ff67bac43b',
        'slovakia': 'f66a0294491086d4c49dc4a804446e50',
        'slovenia': '6fa3ae3920bcc2c890a0d74435d9d29b',
        'south_africa': 'b7f1412d69922e8551cf91081401ec8d',
        'spain': '908bbf29597077c2c6954c439fe8265f',
        'sweden': '4b07726c421981bb2019e8900023393e',
        'vietnam': '32e1cacebcb2da656d40ab8522eb6737',
    }

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        target: str = '2-class',
        countries: str | Sequence[str] = ['austria'],
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Fields Of The World dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", or "test"
            target: one of "2-class", "3-class", or "instance" specifying which kind of
                target mask to load
            countries: which set of countries to load data from
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: If any arguments are invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits
        assert target in self.targets
        if isinstance(countries, str):
            countries = [countries]
        assert set(countries) <= set(self.valid_countries)

        self.root = root
        self.split = split
        self.target = target
        self.countries = countries
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        self.files = self._load_files()

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            image and mask at that index with image of dimension 3x1024x1024
            and mask of dimension 1024x1024
        """
        win_a_fn = self.files[index]['win_a']
        win_b_fn = self.files[index]['win_b']
        mask_fn = self.files[index]['mask']

        win_a = self._load_image(win_a_fn)
        win_b = self._load_image(win_b_fn)
        mask = self._load_target(mask_fn)

        image = torch.cat((win_a, win_b), dim=0)
        sample = {'image': image, 'mask': mask}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of datapoints in the dataset.

        Returns:
            length of dataset
        """
        return len(self.files)

    def _load_files(self) -> list[dict[str, str]]:
        """Return the paths of the files in the dataset.

        Returns:
            a dictionary with "win_a", "win_b", and "mask" keys containing lists of
            file paths
        """
        files = []
        for country in self.countries:
            df = pd.read_parquet(
                os.path.join(self.root, country, f'chips_{country}.parquet')
            )
            aois = df[df['split'] == self.split]['aoi_id'].values

            for aoi in aois:
                if self.target == 'instance':
                    subdir = 'instance'
                elif self.target == '2-class':
                    subdir = 'semantic_2class'
                elif self.target == '3-class':
                    subdir = 'semantic_3class'

                win_a_fn = os.path.join(
                    self.root, country, 's2_images', 'window_a', f'{aoi}.tif'
                )
                win_b_fn = os.path.join(
                    self.root, country, 's2_images', 'window_b', f'{aoi}.tif'
                )

                # there are 333 AOIs that are missing imagery across the dataset
                if not (os.path.exists(win_a_fn) and os.path.exists(win_b_fn)):
                    continue

                sample = {
                    'win_a': win_a_fn,
                    'win_b': win_b_fn,
                    'mask': os.path.join(
                        self.root, country, 'label_masks', subdir, f'{aoi}.tif'
                    ),
                }
                files.append(sample)

        return files

    def _load_image(self, path: Path) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image

        Returns:
            the loaded image
        """
        filename = os.path.join(path)
        with rasterio.open(filename) as f:
            array: np.typing.NDArray[np.int_] = f.read()
            tensor = array_to_tensor(array).float()
        return tensor

    def _load_target(self, path: Path) -> Tensor:
        """Load a single mask corresponding to image.

        Args:
            path: path to the mask

        Returns:
            the mask of the image
        """
        filename = os.path.join(path)
        with rasterio.open(filename) as f:
            array: np.typing.NDArray[np.int_] = f.read(1)
            tensor = torch.from_numpy(array).long()
        return tensor

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        for country in self.countries:
            if self._verify_data(country):
                continue

            filename = f'{country}.zip'
            pathname = os.path.join(self.root, filename)
            if os.path.exists(pathname):
                extract_archive(pathname, os.path.join(self.root, country))
                continue

            if not self.download:
                raise DatasetNotFoundError(self)

            download_and_extract_archive(
                self.base_url + filename,
                os.path.join(self.root, country),
                filename=filename,
                md5=self.country_to_md5[country] if self.checksum else None,
            )

    def _verify_data(self, country: str) -> bool:
        """Verify that data for a country is extracted.

        Args:
            country: the country to check

        Returns:
            True if the dataset directories and split files are found, else False
        """
        for entry in ['label_masks', 's2_images', f'chips_{country}.parquet']:
            if not os.path.exists(os.path.join(self.root, country, entry)):
                return False

        return True

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample return by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        fig, axs = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))

        win_a = einops.rearrange(sample['image'][0:3], 'c h w -> h w c')
        win_b = einops.rearrange(sample['image'][4:7], 'c h w -> h w c')
        mask = sample['mask']

        win_a = torch.clip(win_a / 3000, 0, 1)
        win_b = torch.clip(win_b / 3000, 0, 1)

        axs[0].imshow(win_a)
        axs[0].set_title('Window A')
        axs[1].imshow(win_b)
        axs[1].set_title('Window B')
        if self.target == 'instance':
            unique_vals = sorted(np.unique(mask))
            for i, val in enumerate(unique_vals):
                mask[mask == val] = i
            bg_mask = mask == 0
            mask = (mask % 9) + 1
            mask[bg_mask] = 0
            axs[2].imshow(mask, vmin=0, vmax=10, cmap='tab10', interpolation='none')
            axs[2].set_title('Instance mask')
        elif self.target == '2-class':
            axs[2].imshow(mask, vmin=0, vmax=2, cmap='gray', interpolation='none')
            axs[2].set_title('2-class mask')
        elif self.target == '3-class':
            axs[2].imshow(mask, vmin=0, vmax=2, cmap='gray', interpolation='none')
            axs[2].set_title('3-class mask')
        for ax in axs:
            ax.axis('off')

        if not show_titles:
            for ax in axs:
                ax.set_title('')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Vaihingen dataset."""

import os
from collections.abc import Callable
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import (
    Path,
    check_integrity,
    draw_semantic_segmentation_masks,
    extract_archive,
    rgb_to_mask,
)


class Vaihingen2D(NonGeoDataset):
    """Vaihingen 2D Semantic Segmentation dataset.

    The `Vaihingen <https://www.isprs.org/education/benchmarks/UrbanSemLab/semantic-labeling.aspx>`__
    dataset is a dataset for urban semantic segmentation used in the 2D Semantic Labeling
    Contest - Vaihingen. This dataset uses the "ISPRS_semantic_labeling_Vaihingen.zip"
    and "ISPRS_semantic_labeling_Vaihingen_ground_truth_COMPLETE.zip" files to create the
    train/test sets used in the challenge. The dataset can be downloaded from
    `here <https://www.isprs.org/education/benchmarks/UrbanSemLab/default.aspx>`__.
    Note, the server contains additional data for 3D Semantic Labeling which
    are currently not supported.

    Dataset format:

    * images are 3-channel RGB geotiffs
    * masks are 3-channel geotiffs with unique RGB values representing the class

    Dataset classes:

    0. Clutter/background
    1. Impervious surfaces
    2. Building
    3. Low Vegetation
    4. Tree
    5. Car

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.5194/isprsannals-I-3-293-2012

    .. versionadded:: 0.2
    """

    filenames = (
        'ISPRS_semantic_labeling_Vaihingen.zip',
        'ISPRS_semantic_labeling_Vaihingen_ground_truth_COMPLETE.zip',
    )
    md5s = ('462b8dca7b6fa9eaf729840f0cdfc7f3', '4802dd6326e2727a352fb735be450277')
    image_root = 'top'
    splits: ClassVar[dict[str, list[str]]] = {
        'train': [
            'top_mosaic_09cm_area1.tif',
            'top_mosaic_09cm_area11.tif',
            'top_mosaic_09cm_area13.tif',
            'top_mosaic_09cm_area15.tif',
            'top_mosaic_09cm_area17.tif',
            'top_mosaic_09cm_area21.tif',
            'top_mosaic_09cm_area23.tif',
            'top_mosaic_09cm_area26.tif',
            'top_mosaic_09cm_area28.tif',
            'top_mosaic_09cm_area3.tif',
            'top_mosaic_09cm_area30.tif',
            'top_mosaic_09cm_area32.tif',
            'top_mosaic_09cm_area34.tif',
            'top_mosaic_09cm_area37.tif',
            'top_mosaic_09cm_area5.tif',
            'top_mosaic_09cm_area7.tif',
        ],
        'test': [
            'top_mosaic_09cm_area6.tif',
            'top_mosaic_09cm_area24.tif',
            'top_mosaic_09cm_area35.tif',
            'top_mosaic_09cm_area16.tif',
            'top_mosaic_09cm_area14.tif',
            'top_mosaic_09cm_area22.tif',
            'top_mosaic_09cm_area10.tif',
            'top_mosaic_09cm_area4.tif',
            'top_mosaic_09cm_area2.tif',
            'top_mosaic_09cm_area20.tif',
            'top_mosaic_09cm_area8.tif',
            'top_mosaic_09cm_area31.tif',
            'top_mosaic_09cm_area33.tif',
            'top_mosaic_09cm_area27.tif',
            'top_mosaic_09cm_area38.tif',
            'top_mosaic_09cm_area12.tif',
            'top_mosaic_09cm_area29.tif',
        ],
    }
    classes = (
        'Clutter/background',
        'Impervious surfaces',
        'Building',
        'Low Vegetation',
        'Tree',
        'Car',
    )
    colormap = (
        (255, 0, 0),
        (255, 255, 255),
        (0, 0, 255),
        (0, 255, 255),
        (0, 255, 0),
        (255, 255, 0),
    )

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Vaihingen2D dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train" or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: If *split* is invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits
        self.root = root
        self.split = split
        self.transforms = transforms
        self.checksum = checksum

        self._verify()

        self.files = []
        for name in self.splits[split]:
            image = os.path.join(root, self.image_root, name)
            mask = os.path.join(root, name)
            if os.path.exists(image) and os.path.exists(mask):
                self.files.append(dict(image=image, mask=mask))

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        image = self._load_image(index)
        mask = self._load_target(index)
        sample = {'image': image, 'mask': mask}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.files)

    def _load_image(self, index: int) -> Tensor:
        """Load a single image.

        Args:
            index: index to return

        Returns:
            the image
        """
        path = self.files[index]['image']
        with Image.open(path) as img:
            array: np.typing.NDArray[np.int_] = np.array(img.convert('RGB'))
            tensor = torch.from_numpy(array)
            # Convert from HxWxC to CxHxW
            tensor = tensor.permute((2, 0, 1)).float()
        return tensor

    def _load_target(self, index: int) -> Tensor:
        """Load the target mask for a single image.

        Args:
            index: index to return

        Returns:
            the target mask
        """
        path = self.files[index]['mask']
        with Image.open(path) as img:
            array: np.typing.NDArray[np.uint8] = np.array(img.convert('RGB'))
            array = rgb_to_mask(array, self.colormap)
            tensor = torch.from_numpy(array)
            # Convert from HxWxC to CxHxW
            tensor = tensor.to(torch.long)
        return tensor

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        if os.path.exists(os.path.join(self.root, self.image_root)):
            return

        # Check if .zip files already exists (if so extract)
        exists = []
        for filename, md5 in zip(self.filenames, self.md5s):
            filepath = os.path.join(self.root, filename)
            if os.path.isfile(filepath):
                if self.checksum and not check_integrity(filepath, md5):
                    raise RuntimeError('Dataset found, but corrupted.')
                exists.append(True)
                extract_archive(filepath)
            else:
                exists.append(False)

        if all(exists):
            return

        raise DatasetNotFoundError(self)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
        alpha: float = 0.5,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle
            alpha: opacity with which to render predictions on top of the imagery

        Returns:
            a matplotlib Figure with the rendered sample
        """
        ncols = 1
        image1 = draw_semantic_segmentation_masks(
            sample['image'][:3], sample['mask'], alpha=alpha, colors=list(self.colormap)
        )
        if 'prediction' in sample:
            ncols += 1
            image2 = draw_semantic_segmentation_masks(
                sample['image'][:3],
                sample['prediction'],
                alpha=alpha,
                colors=list(self.colormap),
            )

        fig, axs = plt.subplots(ncols=ncols, figsize=(ncols * 10, 10))
        if ncols > 1:
            (ax0, ax1) = axs
        else:
            ax0 = axs

        ax0.imshow(image1)
        ax0.axis('off')
        if ncols > 1:
            ax1.imshow(image2)
            ax1.axis('off')

        if show_titles:
            ax0.set_title('Ground Truth')
            if ncols > 1:
                ax1.set_title('Predictions')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Trainers for object detection."""

from functools import partial
from typing import Any

import kornia.augmentation as K
import matplotlib.pyplot as plt
import torch
import torchvision.models.detection
from matplotlib.figure import Figure
from timm.models import adapt_input_conv
from torch import Tensor
from torch.nn.parameter import Parameter
from torchmetrics import MetricCollection
from torchmetrics.detection.mean_ap import MeanAveragePrecision
from torchvision.models import resnet as R
from torchvision.models.detection.backbone_utils import resnet_fpn_backbone
from torchvision.models.detection.retinanet import RetinaNetHead
from torchvision.models.detection.rpn import AnchorGenerator
from torchvision.ops import MultiScaleRoIAlign, feature_pyramid_network, misc

from ..datasets import RGBBandsMissingError, unbind_samples
from .base import BaseTask

BACKBONE_LAT_DIM_MAP = {
    'resnet18': 512,
    'resnet34': 512,
    'resnet50': 2048,
    'resnet101': 2048,
    'resnet152': 2048,
    'resnext50_32x4d': 2048,
    'resnext101_32x8d': 2048,
    'wide_resnet50_2': 2048,
    'wide_resnet101_2': 2048,
}

BACKBONE_WEIGHT_MAP = {
    'resnet18': R.ResNet18_Weights.DEFAULT,
    'resnet34': R.ResNet34_Weights.DEFAULT,
    'resnet50': R.ResNet50_Weights.DEFAULT,
    'resnet101': R.ResNet101_Weights.DEFAULT,
    'resnet152': R.ResNet152_Weights.DEFAULT,
    'resnext50_32x4d': R.ResNeXt50_32X4D_Weights.DEFAULT,
    'resnext101_32x8d': R.ResNeXt101_32X8D_Weights.DEFAULT,
    'wide_resnet50_2': R.Wide_ResNet50_2_Weights.DEFAULT,
    'wide_resnet101_2': R.Wide_ResNet101_2_Weights.DEFAULT,
}


class ObjectDetectionTask(BaseTask):
    """Object detection.

    .. versionadded:: 0.4
    """

    ignore = None
    monitor = 'val_map'
    mode = 'max'

    def __init__(
        self,
        model: str = 'faster-rcnn',
        backbone: str = 'resnet50',
        weights: bool | None = None,
        in_channels: int = 3,
        num_classes: int = 1000,
        trainable_layers: int = 3,
        lr: float = 1e-3,
        patience: int = 10,
        freeze_backbone: bool = False,
    ) -> None:
        """Initialize a new ObjectDetectionTask instance.

        Args:
            model: Name of the `torchvision
                <https://pytorch.org/vision/stable/models.html#object-detection>`__
                model to use. One of 'faster-rcnn', 'fcos', or 'retinanet'.
            backbone: Name of the `torchvision
                <https://pytorch.org/vision/stable/models.html#classification>`__
                backbone to use. One of 'resnet18', 'resnet34', 'resnet50',
                'resnet101', 'resnet152', 'resnext50_32x4d', 'resnext101_32x8d',
                'wide_resnet50_2', or 'wide_resnet101_2'.
            weights: Initial model weights. True for ImageNet weights, False or None
                for random weights.
            in_channels: Number of input channels to model.
            num_classes: Number of prediction classes (including the background).
            trainable_layers: Number of trainable layers.
            lr: Learning rate for optimizer.
            patience: Patience for learning rate scheduler.
            freeze_backbone: Freeze the backbone network to fine-tune the detection
                head.

        .. versionchanged:: 0.4
           *detection_model* was renamed to *model*.

        .. versionadded:: 0.5
           The *freeze_backbone* parameter.

        .. versionchanged:: 0.5
           *pretrained*, *learning_rate*, and *learning_rate_schedule_patience* were
           renamed to *weights*, *lr*, and *patience*.
        """
        super().__init__()

    def configure_models(self) -> None:
        """Initialize the model.

        Raises:
            ValueError: If *model* or *backbone* are invalid.
        """
        backbone: str = self.hparams['backbone']
        model: str = self.hparams['model']
        weights: bool | None = self.hparams['weights']
        in_channels: int = self.hparams['in_channels']
        num_classes: int = self.hparams['num_classes']
        freeze_backbone: bool = self.hparams['freeze_backbone']

        if backbone in BACKBONE_LAT_DIM_MAP:
            kwargs = {
                'backbone_name': backbone,
                'trainable_layers': self.hparams['trainable_layers'],
            }
            if weights:
                kwargs['weights'] = BACKBONE_WEIGHT_MAP[backbone]
            else:
                kwargs['weights'] = None

            latent_dim = BACKBONE_LAT_DIM_MAP[backbone]
        else:
            raise ValueError(f"Backbone type '{backbone}' is not valid.")

        if model == 'faster-rcnn':
            model_backbone = resnet_fpn_backbone(**kwargs)
            anchor_generator = AnchorGenerator(
                sizes=((32), (64), (128), (256), (512)), aspect_ratios=((0.5, 1.0, 2.0))
            )

            roi_pooler = MultiScaleRoIAlign(
                featmap_names=['0', '1', '2', '3'], output_size=7, sampling_ratio=2
            )

            if freeze_backbone:
                for param in model_backbone.parameters():
                    param.requires_grad = False

            self.model = torchvision.models.detection.FasterRCNN(
                model_backbone,
                num_classes,
                rpn_anchor_generator=anchor_generator,
                box_roi_pool=roi_pooler,
            )
        elif model == 'fcos':
            kwargs['extra_blocks'] = feature_pyramid_network.LastLevelP6P7(256, 256)
            kwargs['norm_layer'] = (
                misc.FrozenBatchNorm2d if weights else torch.nn.BatchNorm2d
            )

            model_backbone = resnet_fpn_backbone(**kwargs)
            anchor_generator = AnchorGenerator(
                sizes=((8,), (16,), (32,), (64,), (128,), (256,)),
                aspect_ratios=((1.0,), (1.0,), (1.0,), (1.0,), (1.0,), (1.0,)),
            )

            if freeze_backbone:
                for param in model_backbone.parameters():
                    param.requires_grad = False

            self.model = torchvision.models.detection.FCOS(
                model_backbone, num_classes, anchor_generator=anchor_generator
            )
        elif model == 'retinanet':
            kwargs['extra_blocks'] = feature_pyramid_network.LastLevelP6P7(
                latent_dim, 256
            )
            model_backbone = resnet_fpn_backbone(**kwargs)

            anchor_sizes = (
                (16, 20, 25),
                (32, 40, 50),
                (64, 80, 101),
                (128, 161, 203),
                (256, 322, 406),
                (512, 645, 812),
            )
            aspect_ratios = ((0.5, 1.0, 2.0),) * len(anchor_sizes)
            anchor_generator = AnchorGenerator(anchor_sizes, aspect_ratios)

            head = RetinaNetHead(
                model_backbone.out_channels,
                anchor_generator.num_anchors_per_location()[0],
                num_classes,
                norm_layer=partial(torch.nn.GroupNorm, 32),
            )

            if freeze_backbone:
                for param in model_backbone.parameters():
                    param.requires_grad = False

            self.model = torchvision.models.detection.RetinaNet(
                model_backbone,
                num_classes,
                anchor_generator=anchor_generator,
                head=head,
            )
        else:
            raise ValueError(f"Model type '{model}' is not valid.")

        weight = adapt_input_conv(in_channels, self.model.backbone.body.conv1.weight)
        self.model.backbone.body.conv1.weight = Parameter(weight)
        self.model.backbone.body.conv1.in_channels = in_channels

    def configure_metrics(self) -> None:
        """Initialize the performance metrics.

        * :class:`~torchmetrics.detection.mean_ap.MeanAveragePrecision`: Mean average
          precision (mAP) and mean average recall (mAR). Precision is the number of
          true positives divided by the number of true positives + false positives.
          Recall is the number of true positives divived by the number of true positives
          + false negatives. Uses 'macro' averaging. Higher values are better.

        .. note::
           * 'Micro' averaging suits overall performance evaluation but may not
             reflect minority class accuracy.
           * 'Macro' averaging gives equal weight to each class, and is useful for
             balanced performance assessment across imbalanced classes.
        """
        metrics = MetricCollection([MeanAveragePrecision(average='macro')])
        self.val_metrics = metrics.clone(prefix='val_')
        self.test_metrics = metrics.clone(prefix='test_')

    def training_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """Compute the training loss.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            The loss tensor.
        """
        x = batch['image']
        batch_size = x.shape[0]
        assert 'bbox_xyxy' in batch, 'bbox_xyxy is required for object detection.'
        y = [
            {'boxes': batch['bbox_xyxy'][i], 'labels': batch['label'][i]}
            for i in range(batch_size)
        ]
        loss_dict = self(x, y)
        train_loss: Tensor = sum(loss_dict.values())
        self.log_dict(loss_dict, batch_size=batch_size)
        return train_loss

    def validation_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """Compute the validation metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch['image']
        batch_size = x.shape[0]
        assert 'bbox_xyxy' in batch, 'bbox_xyxy is required for object detection.'
        y = [
            {'boxes': batch['bbox_xyxy'][i], 'labels': batch['label'][i]}
            for i in range(batch_size)
        ]
        y_hat = self(x)
        metrics = self.val_metrics(y_hat, y)

        # https://github.com/Lightning-AI/torchmetrics/pull/1832#issuecomment-1623890714
        metrics.pop('val_classes', None)

        self.log_dict(metrics, batch_size=batch_size)

        if (
            batch_idx < 10
            and hasattr(self.trainer, 'datamodule')
            and hasattr(self.trainer.datamodule, 'plot')
            and self.logger
            and hasattr(self.logger, 'experiment')
            and hasattr(self.logger.experiment, 'add_figure')
        ):
            datamodule = self.trainer.datamodule
            aug = K.AugmentationSequential(
                K.Denormalize(datamodule.mean, datamodule.std),
                data_keys=None,
                keepdim=True,
            )
            batch = aug(batch)
            batch['prediction_bbox_xyxy'] = [b['boxes'].cpu() for b in y_hat]
            batch['prediction_label'] = [b['labels'].cpu() for b in y_hat]
            batch['prediction_score'] = [b['scores'].cpu() for b in y_hat]
            batch['image'] = batch['image'].cpu()
            sample = unbind_samples(batch)[0]
            # Convert image to uint8 for plotting
            if torch.is_floating_point(sample['image']):
                sample['image'] *= 255
                sample['image'] = sample['image'].to(torch.uint8)

            fig: Figure | None = None
            try:
                fig = datamodule.plot(sample)
            except RGBBandsMissingError:
                pass

            if fig:
                summary_writer = self.logger.experiment
                summary_writer.add_figure(
                    f'image/{batch_idx}', fig, global_step=self.global_step
                )
                plt.close()

    def test_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        """Compute the test metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch['image']
        batch_size = x.shape[0]
        assert 'bbox_xyxy' in batch, 'bbox_xyxy is required for object detection.'
        y = [
            {'boxes': batch['bbox_xyxy'][i], 'labels': batch['label'][i]}
            for i in range(batch_size)
        ]
        y_hat = self(x)
        metrics = self.test_metrics(y_hat, y)

        # https://github.com/Lightning-AI/torchmetrics/pull/1832#issuecomment-1623890714
        metrics.pop('test_classes', None)

        self.log_dict(metrics, batch_size=batch_size)

    def predict_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> list[dict[str, Tensor]]:
        """Compute the predicted bounding boxes.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """
        x = batch['image']
        y_hat: list[dict[str, Tensor]] = self(x)
        return y_hat



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Self-Supervised Learning for Earth Observation."""

import glob
import os
import random
import re
from collections.abc import Callable
from typing import ClassVar, Literal, TypedDict

import matplotlib.pyplot as plt
import numpy as np
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .landsat import Landsat, Landsat5TM, Landsat7, Landsat8
from .sentinel import Sentinel1, Sentinel2
from .utils import Path, disambiguate_timestamp, download_url, extract_archive


class SSL4EO(NonGeoDataset):
    """Base class for all SSL4EO datasets.

    Self-Supervised Learning for Earth Observation (SSL4EO) is a collection of
    large-scale multimodal multitemporal datasets for unsupervised/self-supervised
    pre-training in Earth observation.

    .. versionadded:: 0.5
    """


class SSL4EOL(SSL4EO):
    """SSL4EO-L dataset.

    Landsat version of SSL4EO.

    The dataset consists of a parallel corpus (same locations and dates for SR/TOA)
    for the following sensors:

    .. list-table::
       :widths: 10 10 10 10 10 10
       :header-rows: 1

       * - Split
         - Satellites
         - Sensors
         - Level
         - # Bands
         - Link
       * - tm_toa
         - Landsat 4--5
         - TM
         - TOA
         - 7
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LT05_C02_T1_TOA>`__
       * - etm_sr
         - Landsat 7
         - ETM+
         - SR
         - 6
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LT05_C02_T1_L2>`__
       * - etm_toa
         - Landsat 7
         - ETM+
         - TOA
         - 9
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LE07_C02_T1_TOA>`__
       * - oli_tirs_toa
         - Landsat 8--9
         - OLI+TIRS
         - TOA
         - 11
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LC08_C02_T1_TOA>`__
       * - oli_sr
         - Landsat 8--9
         - OLI
         - SR
         - 7
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LC08_C02_T1_L2>`__

    Each patch has the following properties:

    * 264 x 264 pixels
    * Resampled to 30 m resolution (7920 x 7920 m)
    * 4 seasonal timestamps
    * Single multispectral GeoTIFF file

    .. note::

       Each split is 300--400 GB and requires 3x that to concatenate and extract
       tarballs. Tarballs can be safely deleted after extraction to save space.
       The dataset takes about 1.5 hrs to download and checksum and another 3 hrs
       to extract.

    If you use this dataset in your research, please cite the following paper:

    * https://proceedings.neurips.cc/paper_files/paper/2023/hash/bbf7ee04e2aefec136ecf60e346c2e61-Abstract-Datasets_and_Benchmarks.html

    .. versionadded:: 0.5
    """

    class _Metadata(TypedDict):
        all_bands: list[str]
        rgb_bands: list[int]

    metadata: ClassVar[dict[str, _Metadata]] = {
        'tm_toa': {
            'all_bands': ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7'],
            'rgb_bands': [2, 1, 0],
        },
        'etm_toa': {
            'all_bands': ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B6', 'B7', 'B8'],
            'rgb_bands': [2, 1, 0],
        },
        'etm_sr': {
            'all_bands': ['B1', 'B2', 'B3', 'B4', 'B5', 'B7'],
            'rgb_bands': [2, 1, 0],
        },
        'oli_tirs_toa': {
            'all_bands': [
                'B1',
                'B2',
                'B3',
                'B4',
                'B5',
                'B6',
                'B7',
                'B8',
                'B9',
                'B10',
                'B11',
            ],
            'rgb_bands': [3, 2, 1],
        },
        'oli_sr': {
            'all_bands': ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7'],
            'rgb_bands': [3, 2, 1],
        },
    }

    url = 'https://hf.co/datasets/torchgeo/ssl4eo_l/resolve/e2467887e6a6bcd7547d9d5999f8d9bc3323dc31/{0}/ssl4eo_l_{0}.tar.gz{1}'
    checksums: ClassVar[dict[str, dict[str, str]]] = {
        'tm_toa': {
            'aa': '553795b8d73aa253445b1e67c5b81f11',
            'ab': 'e9e0739b5171b37d16086cb89ab370e8',
            'ac': '6cb27189f6abe500c67343bfcab2432c',
            'ad': '15a885d4f544d0c1849523f689e27402',
            'ae': '35523336bf9f8132f38ff86413dcd6dc',
            'af': 'fa1108436034e6222d153586861f663b',
            'ag': 'd5c91301c115c00acaf01ceb3b78c0fe',
        },
        'etm_toa': {
            'aa': '587c3efc7d0a0c493dfb36139d91ccdf',
            'ab': 'ec34f33face893d2d8fd152496e1df05',
            'ac': '947acc2c6bc3c1d1415ac92bab695380',
            'ad': 'e31273dec921e187f5c0dc73af5b6102',
            'ae': '43390a47d138593095e9a6775ae7dc75',
            'af': '082881464ca6dcbaa585f72de1ac14fd',
            'ag': 'de2511aaebd640bd5e5404c40d7494cb',
            'ah': '124c5fbcda6871f27524ae59480dabc5',
            'ai': '12b5f94824b7f102df30a63b1139fc57',
        },
        'etm_sr': {
            'aa': 'baa36a9b8e42e234bb44ab4046f8f2ac',
            'ab': '9fb0f948c76154caabe086d2d0008fdf',
            'ac': '99a55367178373805d357a096d68e418',
            'ad': '59d53a643b9e28911246d4609744ef25',
            'ae': '7abfcfc57528cb9c619c66ee307a2cc9',
            'af': 'bb23cf26cc9fe156e7a68589ec69f43e',
            'ag': '97347e5a81d24c93cf33d99bb46a5b91',
        },
        'oli_tirs_toa': {
            'aa': '4711369b861c856ebfadbc861e928d3a',
            'ab': '660a96cda1caf54df837c4b3c6c703f6',
            'ac': 'c9b6a1117916ba318ac3e310447c60dc',
            'ad': 'b8502e9e92d4a7765a287d21d7c9146c',
            'ae': '5c11c14cfe45f78de4f6d6faf03f3146',
            'af': '5b0ed3901be1000137ddd3a6d58d5109',
            'ag': 'a3b6734f8fe6763dcf311c9464a05d5b',
            'ah': '5e55f92e3238a8ab3e471be041f8111b',
            'ai': 'e20617f73d0232a0c0472ce336d4c92f',
        },
        'oli_sr': {
            'aa': 'ca338511c9da4dcbfddda28b38ca9e0a',
            'ab': '7f4100aa9791156958dccf1bb2a88ae0',
            'ac': '6b0f18be2b63ba9da194cc7886dbbc01',
            'ad': '57efbcc894d8da8c4975c29437d8b775',
            'ae': '2594a0a856897f3f5a902c830186872d',
            'af': 'a03839311a2b3dc17dfb9fb9bc4f9751',
            'ag': '6a329d8fd9fdd591e400ab20f9d11dea',
        },
    }

    def __init__(
        self,
        root: Path = 'data',
        split: Literal[
            'tm_toa', 'etm_toa', 'etm_sr', 'oli_tirs_toa', 'oli_sr'
        ] = 'oli_sr',
        seasons: Literal[1, 2, 3, 4] = 1,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new SSL4EOL instance.

        Args:
            root: root directory where dataset can be found
            split: one of ['tm_toa', 'etm_toa', 'etm_sr', 'oli_tirs_toa', 'oli_sr']
            seasons: number of seasonal patches to sample per location, 1--4
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 after downloading files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.root = root
        self.subdir = os.path.join(root, f'ssl4eo_l_{split}')
        self.split = split
        self.seasons = seasons
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        if split.startswith('tm'):
            base: type[Landsat] = Landsat5TM
        elif split.startswith('etm'):
            base = Landsat7
        else:
            base = Landsat8

        self.wavelengths = []
        for band in self.metadata[split]['all_bands']:
            self.wavelengths.append(base.wavelengths[band])

        self.scenes = sorted(os.listdir(self.subdir))

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            image sample
        """
        root = os.path.join(self.subdir, self.scenes[index])
        subdirs = os.listdir(root)
        subdirs = random.sample(subdirs, self.seasons)

        images = []
        xs = []
        ys = []
        ts = []
        wavelengths = []
        for subdir in subdirs:
            mint, maxt = disambiguate_timestamp(subdir[-8:], Landsat.date_format)
            directory = os.path.join(root, subdir)
            filename = os.path.join(directory, 'all_bands.tif')
            with rasterio.open(filename) as f:
                minx, maxx = f.bounds.left, f.bounds.right
                miny, maxy = f.bounds.bottom, f.bounds.top
                image = f.read()
                images.append(torch.from_numpy(image.astype(np.float32)))
                xs.append((minx + maxx) / 2)
                ys.append((miny + maxy) / 2)
                ts.append((mint + maxt) / 2)
                wavelengths.extend(self.wavelengths)

        sample = {
            'image': torch.cat(images),
            'x': torch.tensor(xs),
            'y': torch.tensor(ys),
            't': torch.tensor(ts),
            'wavelength': torch.tensor(wavelengths),
            'res': torch.tensor(30),
        }

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.scenes)

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        path = os.path.join(self.subdir, '00000*', '*', 'all_bands.tif')
        if glob.glob(path):
            return

        # Check if the tar.gz files have already been downloaded
        exists = []
        for suffix in self.checksums[self.split]:
            path = self.subdir + f'.tar.gz{suffix}'
            exists.append(os.path.exists(path))

        if all(exists):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        for suffix, md5 in self.checksums[self.split].items():
            download_url(
                self.url.format(self.split, suffix),
                self.root,
                md5=md5 if self.checksum else None,
            )

    def _extract(self) -> None:
        """Extract the dataset."""
        # Concatenate all tarballs together
        chunk_size = 2**15  # same as torchvision
        path = self.subdir + '.tar.gz'
        with open(path, 'wb') as f:
            for suffix in self.checksums[self.split]:
                with open(path + suffix, 'rb') as g:
                    while chunk := g.read(chunk_size):
                        f.write(chunk)

        # Extract the concatenated tarball
        extract_archive(path)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        fig, axes = plt.subplots(
            ncols=self.seasons, squeeze=False, figsize=(4 * self.seasons, 4)
        )
        num_bands = len(self.metadata[self.split]['all_bands'])
        rgb_bands = self.metadata[self.split]['rgb_bands']

        for i in range(self.seasons):
            image = sample['image'][i * num_bands : (i + 1) * num_bands].byte()

            image = image[rgb_bands].permute(1, 2, 0)
            axes[0, i].imshow(image)
            axes[0, i].axis('off')

            if show_titles:
                axes[0, i].set_title(f'Split {self.split}, Season {i + 1}')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig


class SSL4EOS12(SSL4EO):
    """SSL4EO-S12 dataset.

    `Sentinel-1/2 <https://github.com/zhu-xlab/SSL4EO-S12>`_ version of SSL4EO.

    The dataset consists of a parallel corpus (same locations and dates)
    for the following satellites:

    .. list-table::
       :widths: 10 10 10 10 10
       :header-rows: 1

       * - Split
         - Satellite
         - Level
         - # Bands
         - Link
       * - s1
         - Sentinel-1
         - GRD
         - 2
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S1_GRD>`__
       * - s2c
         - Sentinel-2
         - TOA
         - 12
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2_HARMONIZED>`__
       * - s2a
         - Sentinel-2
         - SR
         - 13
         - `GEE <https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2_SR_HARMONIZED>`__

    Each patch has the following properties:

    * 264 x 264 pixels
    * Resampled to 10 m resolution (2640 x 2640 m)
    * 4 seasonal timestamps

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2211.07044

    .. note::

       The dataset is about 1.5 TB when compressed and 3.7 TB when uncompressed.

    .. versionadded:: 0.5
    """

    size = 264

    class _Metadata(TypedDict):
        bands: list[str]
        filename_regex: str

    metadata: ClassVar[dict[str, _Metadata]] = {
        's1': {
            'bands': ['VV', 'VH'],
            'filename_regex': r'^.{16}_(?P<date>\d{8}T\d{6})',
        },
        's2c': {
            'bands': [
                'B1',
                'B2',
                'B3',
                'B4',
                'B5',
                'B6',
                'B7',
                'B8',
                'B8A',
                'B9',
                'B10',
                'B11',
                'B12',
            ],
            'filename_regex': r'^(?P<date>\d{8}T\d{6})',
        },
        's2a': {
            'bands': [
                'B1',
                'B2',
                'B3',
                'B4',
                'B5',
                'B6',
                'B7',
                'B8',
                'B8A',
                'B9',
                'B11',
                'B12',
            ],
            'filename_regex': r'^(?P<date>\d{8}T\d{6})',
        },
    }

    url = 'https://hf.co/datasets/wangyi111/SSL4EO-S12/resolve/3f5ddad68ba2ea29d019b0cef6cf292ff8af0d62/{0}/{0}.tar.gz.part{1}'
    filenames: ClassVar[dict[str, str]] = {
        's1': 's1_grd',
        's2c': 's2_l1c',
        's2a': 's2_l2a',
    }
    checksums: ClassVar[dict[str, dict[str, str]]] = {
        's1': {
            'aa': '6df278053fc3e4c3fd7de2f77856e606',
            'ab': '837755b4ba8d82faf254df9e5fec13a7',
            'ac': '6400423305d6084e2006eede75cf288e',
            'ad': '22a50d7362d9cbc9714e0740fe2122c7',
            'ae': 'd6ac97ead00b4296a95376949c946b12',
            'af': 'd8047814061431dc627b9ae345c80891',
            'ag': '089ce0548cb7902ce873181cc61f5d70',
            'ah': '745b48c2896ca764ef54f91e4e7c555e',
            'ai': 'c36595cf9617b3b7ea722f63dcccbedc',
            'aj': 'cf16f1d81e8bff2d663e4eba79ec6fa3',
        },
        's2c': {
            'aa': 'a3ef419cc65c4d8ac19b9acc55726166',
            'ab': '580451e8fdf93067ad79202b95dd1a5c',
            'ac': 'a6f7318868f5ba1d94fb9363b50307e4',
            'ad': '86f324215b04cdf4242d07aaf3cdfe57',
            'ae': '5895a545460f34b1712c17732e0f5533',
            'af': '078078bc58d8ecc214ddfd838f796700',
            'ag': '3557dd4c24a5942020391a5baaf51abb',
            'ah': 'd59f89271e414648663d3acb66121761',
            'ai': '1a213539c989d16da4e5b4e09feaa98a',
            'aj': '0b229af5633c7f63486b6d7771b737db',
            'ak': 'babe8bed884d31b891151f5717a83b5c',
            'al': '8d1f5ad28ee868ab0595c889446b8e5f',
        },
        's2a': {
            'aa': 'ef847d906ab44cc9a94d086a89473833',
            'ab': '4a6a8ed9e2a08887707d83bcb6eb57af',
            'ac': '00b706a771df4c4df4cc70a20d790339',
            'ad': '579024e84bd9ab0b86e1182792c8dcf9',
            'ae': 'e259f3536355b665aea490c22c897e59',
            'af': '2a15be319ad15f749bfd4ed85d14c172',
            'ag': 'd8224cff1e727543473b0111e307110c',
            'ah': '0015d8aa5ea9201e13b401fd61c36c6f',
            'ai': 'dfce87c0a9550177fd4b82887902b6e3',
            'aj': '688392701760b737ad74cb0e8c7fb731',
            'ak': 'd8f3e4b110f22f0477973ed2c35586b6',
            'al': '1cc3641cd52afedaa1c50d14d84a6664',
        },
    }

    def __init__(
        self,
        root: Path = 'data',
        split: Literal['s1', 's2c', 's2a'] = 's2c',
        seasons: Literal[1, 2, 3, 4] = 1,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new SSL4EOS12 instance.

        Args:
            root: root directory where dataset can be found
            split: one of "s1" (Sentinel-1 GRD dual-pol SAR),
                "s2c" (Sentinel-2 Level-1C top-of-atmosphere reflectance), or
                "s2a" (Sentinel-2 Level-2A surface reflectance)
            seasons: number of seasonal patches to sample per location, 1--4
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.

        .. versionadded:: 0.7
           The *download* parameter.
        """
        self.root = root
        self.split = split
        self.seasons = seasons
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        self.bands = self.metadata[self.split]['bands']

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            image sample
        """
        root = os.path.join(self.root, self.split, f'{index:07}')
        subdirs = os.listdir(root)
        subdirs = random.sample(subdirs, self.seasons)
        filename_regex = self.metadata[self.split]['filename_regex']

        images = []
        xs = []
        ys = []
        ts = []
        wavelengths: list[float] = []
        for subdir in subdirs:
            directory = os.path.join(root, subdir)
            if match := re.match(filename_regex, subdir):
                date_str = match.group('date')
                match self.split:
                    case 's1':
                        date_format = Sentinel1.date_format
                    case 's2c' | 's2a':
                        date_format = Sentinel2.date_format
                mint, maxt = disambiguate_timestamp(date_str, date_format)
                for band in self.bands:
                    match self.split:
                        case 's1':
                            wavelengths.append(Sentinel1.wavelength)
                        case 's2c' | 's2a':
                            wavelengths.append(Sentinel2.wavelengths[band])

                    filename = os.path.join(directory, f'{band}.tif')
                    with rasterio.open(filename) as f:
                        minx, maxx = f.bounds.left, f.bounds.right
                        miny, maxy = f.bounds.bottom, f.bounds.top
                        image = f.read(out_shape=(1, self.size, self.size))
                        images.append(torch.from_numpy(image.astype(np.float32)))
                xs.append((minx + maxx) / 2)
                ys.append((miny + maxy) / 2)
                ts.append((mint + maxt) / 2)

        sample = {
            'image': torch.cat(images),
            'x': torch.tensor(xs),
            'y': torch.tensor(ys),
            't': torch.tensor(ts),
            'wavelength': torch.tensor(wavelengths),
            'res': torch.tensor(10),
        }

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return 251079

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        path = os.path.join(self.root, self.split, '00000*', '*', '*.tif')
        if glob.glob(path):
            return

        # Check if the tar.gz files have already been downloaded
        exists = []
        for suffix in self.checksums[self.split]:
            path = os.path.join(
                self.root, self.filenames[self.split] + f'.tar.gz.part{suffix}'
            )
            exists.append(os.path.exists(path))

        if all(exists):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        for suffix, md5 in self.checksums[self.split].items():
            download_url(
                self.url.format(self.filenames[self.split], suffix),
                self.root,
                md5=md5 if self.checksum else None,
            )

    def _extract(self) -> None:
        """Extract the dataset."""
        # Concatenate all tarballs together
        chunk_size = 2**15  # same as torchvision
        path = os.path.join(self.root, self.filenames[self.split] + '.tar.gz')
        with open(path, 'wb') as f:
            for suffix in self.checksums[self.split]:
                with open(f'{path}.part{suffix}', 'rb') as g:
                    while chunk := g.read(chunk_size):
                        f.write(chunk)

        # Extract the concatenated tarball
        extract_archive(path)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        nrows = 2 if self.split == 's1' else 1
        fig, axes = plt.subplots(
            nrows=nrows,
            ncols=self.seasons,
            squeeze=False,
            figsize=(4 * self.seasons, 4 * nrows),
        )

        for i in range(self.seasons):
            image = sample['image'][i * len(self.bands) : (i + 1) * len(self.bands)]

            if self.split == 's1':
                axes[0, i].imshow(image[0])
                axes[1, i].imshow(image[1])
            else:
                image = image[[3, 2, 1]].permute(1, 2, 0)
                image = torch.clamp(image / 3000, min=0, max=1)
                axes[0, i].imshow(image)

            axes[0, i].axis('off')

            if show_titles:
                axes[0, i].set_title(f'Split {self.split}, Season {i + 1}')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench Flood-S1 dataset."""

import glob
import json
import os
import re
from collections.abc import Callable, Sequence
from typing import Literal

import numpy as np
import rasterio as rio
import torch
from matplotlib.colors import ListedColormap
from pyproj import Transformer
from torch import Tensor

from ..utils import Path, disambiguate_timestamp
from .base import CopernicusBenchBase


class CopernicusBenchFloodS1(CopernicusBenchBase):
    """Copernicus-Bench Flood-S1 dataset.

    Flood-S1 is a flood segmentation dataset extracted from a large flood mapping
    dataset Kuro Siwo.

    If you use this dataset in your research, please cite the following papers:

    * https://arxiv.org/abs/2503.11849
    * https://arxiv.org/abs/2311.12056

    .. versionadded:: 0.7
    """

    url = 'https://hf.co/datasets/wangyi111/Copernicus-Bench/resolve/9d252acd3aa0e3da3128e05c6f028647f0e48e5f/l3_flood_s1/flood_s1.zip'
    md5 = 'f4337fee5e90203c6d0c3efeb0b97b8a'
    zipfile = 'flood_s1.zip'
    directory = 'flood_s1'
    filename = 'grid_dict_{}.json'
    filename_regex = r'.{18}_(?P<date>\d{8})'
    date_format = '%Y%m%d'
    all_bands = ('VV', 'VH')
    rgb_bands = ('VV', 'VH')
    cmap = ListedColormap(['black', 'cyan', 'magenta'])
    classes = ('No Water', 'Permanent Waters', 'Floods')

    def __init__(
        self,
        root: Path = 'data',
        split: Literal['train', 'val', 'test'] = 'train',
        mode: Literal[1, 2] = 1,
        bands: Sequence[str] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new CopernicusBenchBase instance.

        Args:
            root: Root directory where dataset can be found.
            split: One of 'train', 'val', or 'test'.
            mode: Number of pre-flood images, 1 or 2.
            bands: Sequence of band names to load (defaults to all bands).
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.root = root
        self.split = split
        self.mode = mode
        self.bands = bands or self.all_bands
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        filepath = os.path.join(root, self.directory, self.filename.format(split))
        with open(filepath) as f:
            self.metadata = json.load(f)
        self.files = sorted(self.metadata.keys())

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and labels at that index.
        """
        key = self.files[index]
        path = self.metadata[key]['path']
        directory = os.path.join(self.root, self.directory, 'data', path)
        mask_path = glob.glob(os.path.join(directory, 'MK0_MLU*.tif'))[0]
        sample = self._load_image(directory) | self._load_mask(mask_path)

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_image(self, path: str) -> dict[str, Tensor]:
        """Load an image and metadata.

        Args:
            path: File path to load.

        Returns:
            An image sample.
        """
        images = []
        times = []
        ptypes = ['SL1', 'MS1']
        if self.mode == 2:
            ptypes.insert(0, 'SL2')

        for ptype in ptypes:
            image = []
            for band in self.bands:
                # Band (every band)
                filepath = glob.glob(os.path.join(path, f'{ptype}_I{band}_*.tif'))[0]
                with rio.open(filepath) as f:
                    image.append(f.read(1).astype(np.float32))

            # Image (every ptype)
            images.append(image)

            # Time (every ptype)
            if match := re.match(self.filename_regex, os.path.basename(filepath)):
                if 'date' in match.groupdict():
                    date_str = match.group('date')
                    mint, maxt = disambiguate_timestamp(date_str, self.date_format)
                    time = (mint + maxt) / 2
                    times.append(time)

        # Location (only once)
        with rio.open(filepath) as f:
            x = (f.bounds.left + f.bounds.right) / 2
            y = (f.bounds.bottom + f.bounds.top) / 2
            transformer = Transformer.from_crs(f.crs, 'epsg:4326', always_xy=True)
            lon, lat = transformer.transform(x, y)

        return {
            'image': torch.tensor(np.array(images)),
            'lat': torch.tensor(lat),
            'lon': torch.tensor(lon),
            'time': torch.tensor(times),
        }



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""CaFFe dataset."""

import glob
import os
import textwrap
from collections.abc import Callable
from typing import ClassVar

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_and_extract_archive, extract_archive


class CaFFe(NonGeoDataset):
    """CaFFe (CAlving Fronts and where to Find thEm) dataset.

    The `CaFFe <https://doi.pangaea.de/10.1594/PANGAEA.940950>`__ dataset is a
    semantic segmentation dataset of marine-terminating glaciers.

    Dataset features:

    * 13,090 train, 2,241 validation, and 3,761 test images
    * varying spatial resolution of 6-20m
    * paired binary calving front segmentation masks
    * paired multi-class land cover segmentation masks

    Dataset format:

    * images are single-channel pngs with dimension 512x512
    * segmentation masks are single-channel pngs

    Dataset classes:

    0. N/A
    1. rock
    2. glacier
    3. ocean/ice melange

    If you use this dataset in your research, please cite the following paper:

    * https://essd.copernicus.org/articles/14/4287/2022/

    .. versionadded:: 0.7
    """

    valid_splits = ('train', 'val', 'test')

    zipfilename = 'caffe.zip'

    data_dir = 'caffe'

    image_dir = 'sar_images'

    mask_dirs = ('fronts', 'zones')

    url = 'https://huggingface.co/datasets/torchgeo/caffe/resolve/cc96e8418981ce0f03afc9beace6422fdd7142c4/caffe.zip'

    md5 = '9a92fd6f05af74fbc41602595a55df0d'

    px_class_values_zones: ClassVar[dict[int, str]] = {
        0: 'N/A',
        64: 'rock',
        127: 'glacier',
        254: 'ocean/ice melange',
    }

    zone_class_colors = ('black', 'brown', 'lightgray', 'blue')
    zone_cmap = ListedColormap(zone_class_colors)

    px_class_values_fronts: ClassVar[dict[int, str]] = {0: 'no front', 255: 'front'}

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new instance of CaFFe dataset.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` argument is invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.valid_splits, f'split must be one of {self.valid_splits}'

        self.root = root
        self.split = split
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        self.fpaths = glob.glob(
            os.path.join(
                self.root,
                self.zipfilename.replace('.zip', ''),
                self.mask_dirs[1],
                self.split,
                '*.png',
            )
        )

        self.ordinal_map_zones = torch.zeros(
            max(self.px_class_values_zones.keys()) + 1, dtype=torch.long
        )
        for ordinal, px_class in enumerate(self.px_class_values_zones.keys()):
            self.ordinal_map_zones[px_class] = ordinal

        self.ordinal_map_fronts = torch.zeros(
            max(self.px_class_values_fronts.keys()) + 1, dtype=torch.long
        )
        for ordinal, px_class in enumerate(self.px_class_values_fronts.keys()):
            self.ordinal_map_fronts[px_class] = ordinal

    def __len__(self) -> int:
        """Return the number of images in the dataset."""
        return len(self.fpaths)

    def __getitem__(self, idx: int) -> dict[str, Tensor]:
        """Return the image and mask at the given index.

        Args:
            idx: index of the image and mask to return

        Returns:
            dict: a dict containing the image and mask
        """
        zones_filename = os.path.basename(self.fpaths[idx])
        img_filename = zones_filename.replace('_zones_', '_')
        front_filename = zones_filename.replace('_zones_', '_front_')

        def read_tensor(path: str) -> Tensor:
            return torch.from_numpy(np.array(Image.open(path)))

        img_path = os.path.join(
            self.root, self.data_dir, self.image_dir, self.split, img_filename
        )
        img = read_tensor(img_path).unsqueeze(0).float()

        front_mask = read_tensor(
            os.path.join(
                self.root, self.data_dir, self.mask_dirs[0], self.split, front_filename
            )
        ).long()

        zone_mask = read_tensor(
            os.path.join(
                self.root, self.data_dir, self.mask_dirs[1], self.split, zones_filename
            )
        ).long()

        zone_mask = self.ordinal_map_zones[zone_mask]
        front_mask = self.ordinal_map_fronts[front_mask]

        sample = {'image': img, 'mask_front': front_mask, 'mask_zones': zone_mask}

        if self.transforms:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        exists = []
        if os.path.exists(
            os.path.join(
                self.root,
                self.zipfilename.replace('.zip', ''),
                self.image_dir,
                self.split,
            )
        ):
            exists.append(True)
        else:
            exists.append(False)

        for mask_dir in self.mask_dirs:
            if os.path.exists(
                os.path.join(
                    self.root,
                    self.zipfilename.replace('.zip', ''),
                    mask_dir,
                    self.split,
                )
            ):
                exists.append(True)
            else:
                exists.append(False)

        if all(exists):
            return

        # check download of zipfile
        if os.path.exists(os.path.join(self.root, self.zipfilename)):
            self._extract()
            return

        if not self.download:
            raise DatasetNotFoundError(self)

        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        download_and_extract_archive(
            self.url,
            self.root,
            filename=self.zipfilename,
            md5=self.md5 if self.checksum else None,
        )

    def _extract(self) -> None:
        """Extract the dataset."""
        extract_archive(os.path.join(self.root, self.zipfilename), self.root)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`CaFFe.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        if 'prediction' in sample:
            ncols = 4
        else:
            ncols = 3
        fig, axs = plt.subplots(1, ncols, figsize=(15, 5))

        axs[0].imshow(sample['image'].permute(1, 2, 0).numpy())
        axs[0].axis('off')

        axs[1].imshow(sample['mask_front'].numpy(), cmap='gray')
        axs[1].axis('off')

        unique_classes = np.unique(sample['mask_zones'].numpy())
        axs[2].imshow(sample['mask_zones'].numpy(), cmap=self.zone_cmap)
        axs[2].axis('off')

        handles = [
            mpatches.Patch(
                color=self.zone_cmap(ordinal),
                label='\n'.join(
                    textwrap.wrap(self.px_class_values_zones[px_class], width=10)
                ),
            )
            for ordinal, px_class in enumerate(self.px_class_values_zones.keys())
            if ordinal in unique_classes
        ]
        axs[2].legend(handles=handles, loc='upper right', bbox_to_anchor=(1.4, 1))

        if show_titles:
            axs[0].set_title('Image')
            axs[1].set_title('Front Mask')
            axs[2].set_title('Zone Mask')

        if 'prediction' in sample:
            axs[3].imshow(sample['prediction'].numpy(), cmap='gray')
            axs[3].axis('off')
            if show_titles:
                axs[3].set_title('Prediction')

        if suptitle:
            fig.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""CV4A Kenya Crop Type dataset."""

import os
from collections.abc import Callable, Sequence
from functools import lru_cache

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError, RGBBandsMissingError
from .geo import NonGeoDataset
from .utils import Path, which


class CV4AKenyaCropType(NonGeoDataset):
    """CV4A Kenya Crop Type Competition dataset.

    The `CV4A Kenya Crop Type Competition
    <https://beta.source.coop/repositories/radiantearth/african-crops-kenya-02/>`__
    dataset was produced as part of the Crop Type Detection competition at the
    Computer Vision for Agriculture (CV4A) Workshop at the ICLR 2020 conference.
    The objective of the competition was to create a machine learning model to
    classify fields by crop type from images collected during the growing season
    by the Sentinel-2 satellites.

    See the `dataset documentation
    <https://data.source.coop/radiantearth/african-crops-kenya-02/Documentation.pdf>`__
    for details.

    Consists of 4 tiles of Sentinel 2 imagery from 13 different points in time.

    Each tile has:

    * 13 multi-band observations throughout the growing season. Each observation
      includes 12 bands from Sentinel-2 L2A product, and a cloud probability layer.
      The twelve bands are [B01, B02, B03, B04, B05, B06, B07, B08, B8A,
      B09, B11, B12] (refer to Sentinel-2 documentation for more information about
      the bands). The cloud probability layer is a product of the
      Sentinel-2 atmospheric correction algorithm (Sen2Cor) and provides an estimated
      cloud probability (0-100%) per pixel. All of the bands are mapped to a common
      10 m spatial resolution grid.
    * A raster layer indicating the crop ID for the fields in the training set.
    * A raster layer indicating field IDs for the fields (both training and test sets).
      Fields with a crop ID 0 are the test fields.

    There are 3,286 fields in the train set and 1,402 fields in the test set.

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.34911/RDNT.DW605X

    .. note::

       This dataset requires the following additional library to be installed:

       * `azcopy <https://github.com/Azure/azure-storage-azcopy>`_: to download the
         dataset from Source Cooperative.
    """

    url = 'https://radiantearth.blob.core.windows.net/mlhub/kenya-crop-challenge'
    tiles = tuple(map(str, range(4)))
    dates = (
        '20190606',
        '20190701',
        '20190706',
        '20190711',
        '20190721',
        '20190805',
        '20190815',
        '20190825',
        '20190909',
        '20190919',
        '20190924',
        '20191004',
        '20191103',
    )
    all_bands = (
        'B01',
        'B02',
        'B03',
        'B04',
        'B05',
        'B06',
        'B07',
        'B08',
        'B8A',
        'B09',
        'B11',
        'B12',
        'CLD',
    )
    rgb_bands = ('B04', 'B03', 'B02')

    # Same for all tiles
    tile_height = 3035
    tile_width = 2016

    def __init__(
        self,
        root: Path = 'data',
        chip_size: int = 256,
        stride: int = 128,
        bands: Sequence[str] = all_bands,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
    ) -> None:
        """Initialize a new CV4A Kenya Crop Type Dataset instance.

        Args:
            root: root directory where dataset can be found
            chip_size: size of chips
            stride: spacing between chips, if less than chip_size, then there
                will be overlap between chips
            bands: the subset of bands to load
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory

        Raises:
            AssertionError: If *bands* are invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert set(bands) <= set(self.all_bands)

        self.root = root
        self.chip_size = chip_size
        self.stride = stride
        self.bands = bands
        self.transforms = transforms
        self.download = download

        self._verify()

        # Calculate the indices that we will use over all tiles
        self.chips_metadata = []
        for tile_index in range(len(self.tiles)):
            for y in [
                *list(range(0, self.tile_height - self.chip_size, stride)),
                self.tile_height - self.chip_size,
            ]:
                for x in [
                    *list(range(0, self.tile_width - self.chip_size, stride)),
                    self.tile_width - self.chip_size,
                ]:
                    self.chips_metadata.append((tile_index, y, x))

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data, labels, field ids, and metadata at that index
        """
        tile_index, y, x = self.chips_metadata[index]
        tile = self.tiles[tile_index]

        img = self._load_all_image_tiles(tile)
        labels, field_ids = self._load_label_tile(tile)

        img = img[:, :, y : y + self.chip_size, x : x + self.chip_size]
        labels = labels[y : y + self.chip_size, x : x + self.chip_size]
        field_ids = field_ids[y : y + self.chip_size, x : x + self.chip_size]

        sample = {
            'image': img,
            'mask': labels,
            'field_ids': field_ids,
            'tile_index': torch.tensor(tile_index),
            'x': torch.tensor(x),
            'y': torch.tensor(y),
        }

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of chips in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.chips_metadata)

    @lru_cache(maxsize=128)
    def _load_label_tile(self, tile: str) -> tuple[Tensor, Tensor]:
        """Load a single _tile_ of labels and field_ids.

        Args:
            tile: name of tile to load

        Returns:
            tuple of labels and field ids
        """
        directory = os.path.join(self.root, 'data', tile)

        with Image.open(os.path.join(directory, f'{tile}_label.tif')) as img:
            array: np.typing.NDArray[np.int_] = np.array(img)
            labels = torch.from_numpy(array)

        with Image.open(os.path.join(directory, f'{tile}_field_id.tif')) as img:
            array = np.array(img)
            field_ids = torch.from_numpy(array)

        return labels, field_ids

    @lru_cache(maxsize=128)
    def _load_all_image_tiles(self, tile: str) -> Tensor:
        """Load all the imagery (across time) for a single _tile_.

        Optionally allows for subsetting of the bands that are loaded.

        Args:
            tile: name of tile to load

        Returns:
            imagery of shape (13, number of bands, 3035, 2016) where 13 is the number of
            points in time, 3035 is the tile height, and 2016 is the tile width
        """
        img = torch.zeros(
            len(self.dates),
            len(self.bands),
            self.tile_height,
            self.tile_width,
            dtype=torch.float32,
        )

        for date_index, date in enumerate(self.dates):
            img[date_index] = self._load_single_image_tile(tile, date)

        return img

    @lru_cache(maxsize=128)
    def _load_single_image_tile(self, tile: str, date: str) -> Tensor:
        """Load the imagery for a single tile for a single date.

        Args:
            tile: name of tile to load
            date: date of tile to load

        Returns:
            array containing a single image tile
        """
        directory = os.path.join(self.root, 'data', tile, date)
        img = torch.zeros(
            len(self.bands), self.tile_height, self.tile_width, dtype=torch.float32
        )
        for band_index, band_name in enumerate(self.bands):
            filepath = os.path.join(directory, f'{tile}_{band_name}_{date}.tif')
            with Image.open(filepath) as band_img:
                array: np.typing.NDArray[np.int_] = np.array(band_img)
                img[band_index] = torch.from_numpy(array)

        return img

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        if os.path.exists(os.path.join(self.root, 'FieldIds.csv')):
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        os.makedirs(self.root, exist_ok=True)
        azcopy = which('azcopy')
        azcopy('sync', self.url, self.root, '--recursive=true')

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        time_step: int = 0,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            time_step: time step at which to access image, beginning with 0
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.

        .. versionadded:: 0.2
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        if 'prediction' in sample:
            prediction = sample['prediction']
            n_cols = 3
        else:
            n_cols = 2

        image, mask = sample['image'], sample['mask']

        image = image[time_step, rgb_indices]

        fig, axs = plt.subplots(nrows=1, ncols=n_cols, figsize=(10, n_cols * 5))

        axs[0].imshow(image.permute(1, 2, 0))
        axs[0].axis('off')
        axs[1].imshow(mask)
        axs[1].axis('off')

        if 'prediction' in sample:
            axs[2].imshow(prediction)
            axs[2].axis('off')
            if show_titles:
                axs[2].set_title('Prediction')

        if show_titles:
            axs[0].set_title('Image')
            axs[1].set_title('Mask')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TreeSatAI datasets."""

import json
import os
from collections.abc import Callable, Sequence
from typing import ClassVar

import rasterio as rio
import torch
from einops import rearrange
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, extract_archive, percentile_normalization


class TreeSatAI(NonGeoDataset):
    """TreeSatAI Benchmark Archive.

    `TreeSatAI Benchmark Archive <https://zenodo.org/records/6780578>`_ is a
    multi-sensor, multi-label dataset for tree species classification in remote
    sensing. It was created by combining labels from the federal forest inventory of
    Lower Saxony, Germany with 20 cm Color-Infrared (CIR) and 10 m Sentinel imagery.

    The TreeSatAI Benchmark Archive contains:

    * 50,381 image triplets (aerial, Sentinel-1, Sentinel-2)
    * synchronized time steps and locations
    * all original spectral bands/polarizations from the sensors
    * 20 species classes (single labels)
    * 12 age classes (single labels)
    * 15 genus classes (multi labels)
    * 60 m and 200 m patches
    * fixed split for train (90%) and test (10%) data
    * additional single labels such as English species name, genus,
      forest stand type, foliage type, land cover

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.5194/essd-15-681-2023

    .. versionadded:: 0.7
    """

    url = 'https://zenodo.org/records/6780578/files/'
    md5s: ClassVar[dict[str, str]] = {
        'aerial_60m_abies_alba.zip': '4298b1c9fbf6d0d85f7aa208ff5fe0c9',
        'aerial_60m_acer_pseudoplatanus.zip': '7c31d7ddea841f6509deece8f984a79e',
        'aerial_60m_alnus_spec.zip': '34ea107f43c6172c6d2652dbf26306af',
        'aerial_60m_betula_spec.zip': '69de9373739a027692a823846434fa0c',
        'aerial_60m_cleared.zip': '8dffbb2f6aad17ef83721cffa5b52d96',
        'aerial_60m_fagus_sylvatica.zip': '77b277e69e90bfbd3c5fd15a73d228fe',
        'aerial_60m_fraxinus_excelsior.zip': '9a88a8e6821f8a54ded950de9238831f',
        'aerial_60m_larix_decidua.zip': 'aa0bc5b091b099018a078536ef429031',
        'aerial_60m_larix_kaempferi.zip': '429df073f69f8bbf60aef765e1c925ba',
        'aerial_60m_picea_abies.zip': 'edb9b1bc9a5a7b405f4cbb0d71cedf54',
        'aerial_60m_pinus_nigra.zip': '96bf1798ef82f712ea46c2963ddb7083',
        'aerial_60m_pinus_strobus.zip': '0ff818c6d31f59b8488880e49b300c7a',
        'aerial_60m_pinus_sylvestris.zip': '298cbaac4d9f07a204e1e74e8446798d',
        'aerial_60m_populus_spec.zip': '46fcff76b119cc24f3caf938a0bb433a',
        'aerial_60m_prunus_spec.zip': 'fb1c570d3ea925a049630224ccb354bc',
        'aerial_60m_pseudotsuga_menziesii.zip': '2d05511ceabf4037b869eca928f3c04e',
        'aerial_60m_quercus_petraea.zip': '31f573fb0419b2b453ed7da1c4d2a298',
        'aerial_60m_quercus_robur.zip': 'bcd90506509de26692c043f4c8d73af0',
        'aerial_60m_quercus_rubra.zip': '71d8495725ed1b4f27d9e382409fcc5e',
        'aerial_60m_tilia_spec.zip': 'f81558c9c7189ac8a257d041ee43c1c9',
        'geojson.zip': 'aa749718f3cb76c1dfc9cddc2ed201db',
        'labels.zip': '656f1b68ec9ab70afd02bb127b75bb24',
        's1.zip': 'bed4fc8cb65da46a24ec1bc6cea2763c',
        's2.zip': '453ba69056aa33a3c6b97afb7b6afadb',
        'test_filenames.lst': '2166903d947f0025f61e342da466f917',
        'train_filenames.lst': 'a1a0148e8120b0268f76d2e98a68436f',
    }

    # Genus-level classes (species-level labels also exist)
    classes = (
        'Abies',  # fir
        'Acer',  # maple
        'Alnus',  # alder
        'Betula',  # birch
        'Cleared',  # none
        'Fagus',  # beech
        'Fraxinus',  # ash
        'Larix',  # larch
        'Picea',  # spruce
        'Pinus',  # pine
        'Populus',  # poplar
        'Prunus',  # cherry
        'Pseudotsuga',  # Douglas fir
        'Quercus',  # oak
        'Tilia',  # linden
    )

    # https://zenodo.org/records/6780578/files/220629_doc_TreeSatAI_benchmark_archive.pdf
    all_sensors = ('aerial', 's1', 's2')
    all_bands: ClassVar[dict[str, list[str]]] = {
        'aerial': ['IR', 'G', 'B', 'R'],
        's1': ['VV', 'VH', 'VV/VH'],
        's2': [
            'B02',
            'B03',
            'B04',
            'B08',
            'B05',
            'B06',
            'B07',
            'B8A',
            'B11',
            'B12',
            'B01',
            'B09',
        ],
    }
    rgb_bands: ClassVar[dict[str, list[str]]] = {
        'aerial': ['R', 'G', 'B'],
        's1': ['VV', 'VH', 'VV/VH'],
        's2': ['B04', 'B03', 'B02'],
    }

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        sensors: Sequence[str] = all_sensors,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new TreeSatAI instance.

        Args:
            root: Root directory where dataset can be found.
            split: Either 'train' or 'test'.
            sensors: One or more of 'aerial', 's1', and/or 's2'.
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            AssertionError: If invalid *sensors* are chosen.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert set(sensors) <= set(self.all_sensors)

        self.root = root
        self.split = split
        self.sensors = sensors
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        path = os.path.join(self.root, f'{split}_filenames.lst')
        with open(path) as f:
            self.files = f.read().strip().split('\n')

        path = os.path.join(self.root, 'labels', 'TreeSatBA_v9_60m_multi_labels.json')
        with open(path) as f:
            self.labels = json.load(f)

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            Length of the dataset.
        """
        return len(self.files)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and label at that index.
        """
        file = self.files[index]
        label = torch.zeros(len(self.classes), dtype=torch.long)
        for genus, _ in self.labels[file]:
            i = self.classes.index(genus)
            label[i] = 1

        sample = {'label': label}
        for directory in self.sensors:
            with rio.open(os.path.join(self.root, directory, '60m', file)) as f:
                sample[f'image_{directory}'] = torch.tensor(f.read().astype('float32'))

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        exists = []
        for directory in self.sensors:
            exists.append(os.path.isdir(os.path.join(self.root, directory)))

        if all(exists):
            return

        for file, md5 in self.md5s.items():
            # Check if the file has already been downloaded
            if os.path.isfile(os.path.join(self.root, file)):
                self._extract(file)
                continue

            # Check if the user requested to download the dataset
            if self.download:
                url = self.url + file
                download_url(url, self.root, md5=md5 if self.checksum else None)
                self._extract(file)
                continue

            raise DatasetNotFoundError(self)

    def _extract(self, file: str) -> None:
        """Extract file.

        Args:
            file: The file to extract.
        """
        if not file.endswith('.zip'):
            return

        to_path = self.root
        if file.startswith('aerial'):
            to_path = os.path.join(self.root, 'aerial', '60m')

        extract_archive(os.path.join(self.root, file), to_path)

    def plot(self, sample: dict[str, Tensor], show_titles: bool = True) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: A sample returned by :meth:`__getitem__`.
            show_titles: Flag indicating whether to show titles above each panel.

        Returns:
            A matplotlib Figure with the rendered sample.
        """
        fig, ax = plt.subplots(ncols=len(self.sensors), squeeze=False)

        for i, sensor in enumerate(self.sensors):
            image = sample[f'image_{sensor}'].cpu().numpy()
            bands = [self.all_bands[sensor].index(b) for b in self.rgb_bands[sensor]]
            image = rearrange(image[bands], 'c h w -> h w c')
            image = percentile_normalization(image)
            ax[0, i].imshow(image)
            ax[0, i].axis('off')

            if show_titles:
                ax[0, i].set_title(sensor)

        if show_titles:
            label = self._multilabel_to_string(sample['label'])
            suptitle = f'Label: ({label})'

            if 'prediction' in sample:
                prediction = self._multilabel_to_string(sample['prediction'])
                suptitle += f'\nPrediction: ({prediction})'

            fig.suptitle(suptitle)

        fig.tight_layout()
        return fig

    def _multilabel_to_string(self, multilabel: Tensor) -> str:
        """Convert a tensor of multilabel class probabilities to human readable format.

        Args:
            multilabel: A tensor of multilabel class probabilities.

        Returns:
            Class names and percentages sorted by percentage.
        """
        labels: list[tuple[str, float]] = []
        for i, pct in enumerate(multilabel.cpu().numpy()):
            if pct > 0.001:
                labels.append((self.classes[i], pct))

        labels.sort(key=lambda label: label[1], reverse=True)
        return ', '.join([f'{genus}: {pct:.1%}' for genus, pct in labels])



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Command-line interface to TorchGeo."""

import os

from lightning.pytorch.cli import ArgsType, LightningCLI

# Allows classes to be referenced using only the class name
import torchgeo.datamodules
import torchgeo.trainers  # noqa: F401
from torchgeo.datamodules import BaseDataModule
from torchgeo.trainers import BaseTask


def main(args: ArgsType = None) -> None:
    """Command-line interface to TorchGeo."""
    # Taken from https://github.com/pangeo-data/cog-best-practices
    rasterio_best_practices = {
        'GDAL_DISABLE_READDIR_ON_OPEN': 'EMPTY_DIR',
        'AWS_NO_SIGN_REQUEST': 'YES',
        'GDAL_MAX_RAW_BLOCK_CACHE_SIZE': '200000000',
        'GDAL_SWATH_SIZE': '200000000',
        'VSI_CURL_CACHE_SIZE': '200000000',
    }
    os.environ.update(rasterio_best_practices)

    LightningCLI(
        model_class=BaseTask,
        datamodule_class=BaseDataModule,
        seed_everything_default=0,
        subclass_mode_model=True,
        subclass_mode_data=True,
        save_config_kwargs={'overwrite': True},
        args=args,
    )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SSL4EO datamodule."""

from typing import Any

import torch

from ..datasets import SSL4EOL, SSL4EOS12
from .geo import NonGeoDataModule


class SSL4EOLDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the SSL4EO-L dataset.

    .. versionadded:: 0.5
    """

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new SSL4EOLDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.SSL4EOL`.
        """
        super().__init__(SSL4EOL, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        self.dataset = SSL4EOL(**self.kwargs)


class SSL4EOS12DataModule(NonGeoDataModule):
    """LightningDataModule implementation for the SSL4EO-S12 dataset.

    .. versionadded:: 0.5
    """

    # https://github.com/zhu-xlab/SSL4EO-S12/blob/d2868adfada65e40910bfcedfc49bc3b20df2248/src/benchmark/transfer_classification/datasets/EuroSat/eurosat_dataset.py#L97
    mean = torch.tensor(0)
    std = torch.tensor(10000)

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new SSL4EOS12DataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.SSL4EOS12`.
        """
        super().__init__(SSL4EOS12, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        self.dataset = SSL4EOS12(**self.kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""BigEarthNet datamodule."""

from typing import Any

import torch

from ..datasets import BigEarthNet
from .geo import NonGeoDataModule


class BigEarthNetDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the BigEarthNet dataset.

    Uses the train/val/test splits from the dataset.
    """

    # (VV, VH, B01, B02, B03, B04, B05, B06, B07, B08, B8A, B09, B11, B12)
    # min/max band statistics computed on 100k random samples
    mins_raw = torch.tensor(
        [-70.0, -72.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0]
    )
    maxs_raw = torch.tensor(
        [
            31.0,
            35.0,
            18556.0,
            20528.0,
            18976.0,
            17874.0,
            16611.0,
            16512.0,
            16394.0,
            16672.0,
            16141.0,
            16097.0,
            15336.0,
            15203.0,
        ]
    )

    # min/max band statistics computed by percentile clipping the
    # above to samples to [2, 98]
    mins = torch.tensor(
        [-48.0, -42.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    )
    maxs = torch.tensor(
        [
            6.0,
            16.0,
            9859.0,
            12872.0,
            13163.0,
            14445.0,
            12477.0,
            12563.0,
            12289.0,
            15596.0,
            12183.0,
            9458.0,
            5897.0,
            5544.0,
        ]
    )

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new BigEarthNetDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.BigEarthNet`.
        """
        bands = kwargs.get('bands', 'all')
        if bands == 'all':
            mins = self.mins
            maxs = self.maxs
        elif bands == 's1':
            mins = self.mins[:2]
            maxs = self.maxs[:2]
        else:
            mins = self.mins[2:]
            maxs = self.maxs[2:]
        self.mean = mins
        self.std = maxs - mins

        super().__init__(BigEarthNet, batch_size, num_workers, **kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TreeSatAI datamodules."""

from typing import Any

import kornia.augmentation as K
import torch
from torch import Tensor
from torch.utils.data import random_split

from ..datasets import TreeSatAI
from ..samplers.utils import _to_tuple
from .geo import NonGeoDataModule

# https://git.tu-berlin.de/rsim/treesat_benchmark/-/blob/master/configs/multimodal/AllModes_Xformer_ResnetScratch_v8.json
means = {
    'aerial': [
        151.26809261440323,
        93.1159469148246,
        85.05016794624635,
        81.0471576353153,
    ],
    's1': [-6.933713050794077, -12.628564056094067, 0.47448312147709354],
    's2': [
        231.43385024546893,
        376.94788434611434,
        241.03688288984037,
        2809.8421354087955,
        616.5578221193639,
        2104.3826773960823,
        2695.083864757169,
        2969.868417923599,
        1306.0814241837832,
        587.0608264363341,
        249.1888624097736,
        2950.2294375352285,
    ],
}
stds = {
    'aerial': [
        48.70879149145466,
        33.59622314610158,
        28.000497087051126,
        33.683983599997724,
    ],
    's1': [87.8762246957811, 47.03070478433704, 1.297291303623673],
    's2': [
        123.16515044781909,
        139.78991338362886,
        140.6154081184225,
        786.4508872594147,
        202.51268536579394,
        530.7255451201194,
        710.2650071967689,
        777.4421400779165,
        424.30312334282684,
        247.21468849049668,
        122.80062680549261,
        702.7404237034002,
    ],
}


class TreeSatAIDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the TreeSatAI dataset.

    .. versionadded:: 0.7
    """

    def __init__(
        self,
        batch_size: int = 64,
        patch_size: int | tuple[int, int] = 304,
        num_workers: int = 0,
        **kwargs: Any,
    ) -> None:
        """Initialize a new TreeSatAIDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            patch_size: Size of each patch, either ``size`` or ``(height, width)``.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.TreeSatAI`.
        """
        super().__init__(TreeSatAI, batch_size, num_workers, **kwargs)

        self.patch_size = _to_tuple(patch_size)
        self.sensors = kwargs.get('sensors', TreeSatAI.all_sensors)

        self.train_aug = K.AugmentationSequential(
            K.RandomVerticalFlip(p=0.5),
            K.RandomHorizontalFlip(p=0.5),
            K.Resize(self.patch_size),
            data_keys=None,
            keepdim=True,
        )
        self.aug = K.AugmentationSequential(
            K.Resize(self.patch_size), data_keys=None, keepdim=True
        )

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        # Convert 90-10 train-test split to 80-10-10 train-val-test split
        train_val_dataset = TreeSatAI(split='train', **self.kwargs)
        self.test_dataset = TreeSatAI(split='test', **self.kwargs)
        generator = torch.Generator().manual_seed(0)
        self.train_dataset, self.val_dataset = random_split(
            train_val_dataset,
            [len(train_val_dataset) - len(self.test_dataset), len(self.test_dataset)],
            generator=generator,
        )

    def on_after_batch_transfer(
        self, batch: dict[str, Tensor], dataloader_idx: int
    ) -> dict[str, Tensor]:
        """Apply batch augmentations to the batch after it is transferred to the device.

        Args:
            batch: A batch of data that needs to be altered or augmented.
            dataloader_idx: The index of the dataloader to which the batch belongs.

        Returns:
            A batch of data.
        """
        batch = super().on_after_batch_transfer(batch, dataloader_idx)

        images = []
        for sensor in self.sensors:
            aug = K.Normalize(mean=means[sensor], std=stds[sensor], keepdim=True)
            batch[f'image_{sensor}'] = aug(batch[f'image_{sensor}'])
            images.append(batch[f'image_{sensor}'])

        batch['image'] = torch.cat(images, dim=1)

        return batch



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Esri 2020 Land Cover Dataset."""

import glob
import os
from collections.abc import Callable, Iterable
from typing import Any

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import DatasetNotFoundError
from .geo import RasterDataset
from .utils import Path, download_url, extract_archive


class Esri2020(RasterDataset):
    """Esri 2020 Land Cover Dataset.

    The `Esri 2020 Land Cover dataset
    <https://www.arcgis.com/home/item.html?id=fc92d38533d440078f17678ebc20e8e2>`_
    consists of a global single band land use/land cover map derived from ESA
    Sentinel-2 imagery at 10m resolution with a total of 10 classes.
    It was published in July 2021 and used the Universal Transverse Mercator (UTM)
    projection. This dataset only contains labels, no raw satellite imagery.

    The 10 classes are:

    0. No Data
    1. Water
    2. Trees
    3. Grass
    4. Flooded Vegetation
    5. Crops
    6. Scrub/Shrub
    7. Built Area
    8. Bare Ground
    9. Snow/Ice
    10. Clouds

    A more detailed explanation of the individual classes can be found
    `here <https://www.arcgis.com/home/item.html?id=fc92d38533d440078f17678ebc20e8e2>`_.

    If you use this dataset please cite the following paper:

    * https://ieeexplore.ieee.org/document/9553499

    .. versionadded:: 0.3
    """

    is_image = False
    filename_glob = '*_20200101-20210101.*'
    filename_regex = r"""^
        (?P<id>[0-9][0-9][A-Z])
        _(?P<date>\d{8})
        -(?P<processing_date>\d{8})
    """

    zipfile = 'io-lulc-model-001-v01-composite-v03-supercell-v02-clip-v01.zip'
    md5 = '4932855fcd00735a34b74b1f87db3df0'

    url = (
        'https://ai4edataeuwest.blob.core.windows.net/io-lulc/'
        'io-lulc-model-001-v01-composite-v03-supercell-v02-clip-v01.zip'
    )

    def __init__(
        self,
        paths: Path | Iterable[Path] = 'data',
        crs: CRS | None = None,
        res: float | tuple[float, float] | None = None,
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        cache: bool = True,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            paths: one or more root directories to search or files to load
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS in (xres, yres) format. If a
                single float is provided, it is used for both the x and y resolution.
                (defaults to the resolution of the first file found)
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.

        .. versionchanged:: 0.5
           *root* was renamed to *paths*.
        """
        self.paths = paths
        self.download = download
        self.checksum = checksum

        self._verify()

        super().__init__(paths, crs, res, transforms=transforms, cache=cache)

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted file already exists
        if self.files:
            return

        # Check if the zip files have already been downloaded
        assert isinstance(self.paths, str | os.PathLike)
        pathname = os.path.join(self.paths, self.zipfile)
        if glob.glob(pathname):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        download_url(self.url, self.paths, filename=self.zipfile, md5=self.md5)

    def _extract(self) -> None:
        """Extract the dataset."""
        assert isinstance(self.paths, str | os.PathLike)
        extract_archive(os.path.join(self.paths, self.zipfile))

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        mask = sample['mask'].squeeze()
        ncols = 1

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = sample['prediction'].squeeze()
            ncols = 2

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(4 * ncols, 4))

        if showing_predictions:
            axs[0].imshow(mask)
            axs[0].axis('off')
            axs[1].imshow(prediction)
            axs[1].axis('off')
            if show_titles:
                axs[0].set_title('Mask')
                axs[1].set_title('Prediction')
        else:
            axs.imshow(mask)
            axs.axis('off')
            if show_titles:
                axs.set_title('Mask')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet 1 dataset."""

from typing import ClassVar

from .base import SpaceNet


class SpaceNet1(SpaceNet):
    """SpaceNet 1: Building Detection v1 Dataset.

    `SpaceNet 1 <https://spacenet.ai/spacenet-buildings-dataset-v1/>`_
    is a dataset of building footprints over the city of Rio de Janeiro.

    Dataset features:

    * No. of images: 6940 (8 Band) + 6940 (RGB)
    * No. of polygons: 382,534 building labels
    * Area Coverage: 2544 sq km
    * GSD: 1 m (8 band),  50 cm (rgb)
    * Chip size: 102 x 110 (8 band), 406 x 439 (rgb)

    Dataset format:

    * Imagery - Worldview-2 GeoTIFFs

        * 8Band.tif (Multispectral)
        * RGB.tif (Pansharpened RGB)

    * Labels - GeoJSON

        * labels.geojson

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/1807.01232
    """

    directory_glob = '{product}'
    dataset_id = 'SN1_buildings'
    tarballs: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            1: [
                'SN1_buildings_train_AOI_1_Rio_3band.tar.gz',
                'SN1_buildings_train_AOI_1_Rio_8band.tar.gz',
                'SN1_buildings_train_AOI_1_Rio_geojson_buildings.tar.gz',
            ]
        },
        'test': {
            1: [
                'SN1_buildings_test_AOI_1_Rio_3band.tar.gz',
                'SN1_buildings_test_AOI_1_Rio_8band.tar.gz',
            ]
        },
    }
    md5s: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            1: [
                '279e334a2120ecac70439ea246174516',
                '6440a9eedbd7c4fe9741875135362c8c',
                'b6e02fbd727f252ea038abe4f77a77b3',
            ]
        },
        'test': {
            1: ['18283d78b21c239bc1831f3bf1d2c996', '732b3a40603b76e80aac84e002e2b3e8']
        },
    }
    valid_aois: ClassVar[dict[str, list[int]]] = {'train': [1], 'test': [1]}
    valid_images: ClassVar[dict[str, list[str]]] = {
        'train': ['3band', '8band'],
        'test': ['3band', '8band'],
    }
    valid_masks = ('geojson',)
    chip_size: ClassVar[dict[str, tuple[int, int]]] = {
        '3band': (406, 439),
        '8band': (102, 110),
    }
